"""
Provide helper functions for constructing sub-commands
"""

import scanpy as sc
import pandas as pd

sc_warn = sc.logging.warn if sc.__version__.startswith("1.4") else sc.logging.warning


def write_cluster(adata, keys, cluster_fn, sep="\t"):
    """Export cell clustering as a text table"""
    if not isinstance(keys, (list, tuple)):
        keys = [keys]
    for key in keys:
        if key not in adata.obs.keys():
            raise KeyError(f"{key} is not a valid `.uns` key")
    adata.obs[keys].reset_index(level=0).rename(columns={"index": "cells"}).to_csv(
        cluster_fn, sep=sep, header=True, index=False
    )


def write_embedding(adata, key, embed_fn, n_comp=None, sep="\t", key_added=None):
    """Export cell embeddings as a txt table"""
    if key_added:
        if embed_fn.endswith(".tsv"):
            embed_fn = embed_fn[0:-4]
        embed_fn = f"{embed_fn}_{key_added}.tsv"
    if key not in adata.obsm.keys():
        raise KeyError(f"{key} is not a valid `.obsm` key")
    mat = adata.obsm[key].copy()
    if n_comp is not None and mat.shape[1] >= n_comp:
        mat = mat[:, 0:n_comp]
    pd.DataFrame(mat, index=adata.obs_names).to_csv(embed_fn, sep=sep, header=False, index=True)


# The functions below handles slot key.
#
# Default keys are those read and written by scanpy functions by default, e.g
# "X_pca", "neighbors", "louvain", etc.
#
# Of them, `obsm_key` specifically refers to those used for embedding, e.g
# "X_pca", "X_tsne", "X_umap", etc.
#
# The approach for supplying a non-standard key to a function as input is:
# if the function only reads the value in the default key, we first backup the
# value in the default key, then write the value of the non-standard key into
# the standard key, run the funtion, and finally restore the value of the
# default key from backup and delete the backup.
#
# The approach for writting the results of a function to a non-standard key is:
# if the function only writes to the default key, we first backup the value in
# the default key, run the function, copy the value of the default key to the
# desired non-standard key, and finally restore the value of the default key
# from backup and delete the backup.
#
# Specical treatment for obsm_key is needed, as the underlying data type is not
# a python dictionary but a numpy array.


def _backup_default_key(slot, default):
    if default in slot.keys():
        bkup_key = f"{default}_bkup"
        if bkup_key in slot.keys():
            sc_warn(f"overwrite existing {bkup_key}")
        slot[bkup_key] = slot[default]


def _restore_default_key(slot, default, key=None):
    if key != default:
        bkup_key = f"{default}_bkup"
        if bkup_key in slot.keys():
            slot[default] = slot[bkup_key]
            del slot[bkup_key]


def _delete_backup_key(slot, default):
    bkup_key = f"{default}_bkup"
    if bkup_key in slot.keys():
        del slot[bkup_key]


def _set_default_key(slot, default, key):
    if key != default:
        if key not in slot.keys():
            raise KeyError(f"{key} does not exist")
        _backup_default_key(slot, default)
        slot[default] = slot[key]


def _rename_default_key(slot, default, key):
    if not default in slot.keys():
        raise KeyError(f"{default} does not exist")
    slot[key] = slot[default]
    del slot[default]
    _restore_default_key(slot, default)


def _backup_obsm_key(adata, key):
    if key in adata.obsm_keys():
        bkup_key = f"{key}_bkup"
        if bkup_key in adata.obsm_keys():
            sc_warn(f"overwrite existing {bkup_key}")
        adata.obsm[bkup_key] = adata.obsm[key]


def _restore_obsm_key(adata, key, new_key=None):
    if new_key != key:
        bkup_key = f"{key}_bkup"
        if bkup_key in adata.obsm_keys():
            adata.obsm[key] = adata.obsm[bkup_key]
            del adata.obsm[bkup_key]


def _delete_obsm_backup_key(adata, key):
    bkup_key = f"{key}_bkup"
    if bkup_key in adata.obsm_keys():
        del adata.obsm[bkup_key]


def _set_obsm_key(adata, key, new_key):
    if new_key != key:
        if new_key not in adata.obsm_keys():
            raise KeyError(f"{new_key} does not exist")
        _backup_obsm_key(adata, key)
        adata.obsm[key] = adata.obsm[new_key]


def _rename_obsm_key(adata, from_key, to_key):
    if not from_key in adata.obsm_keys():
        raise KeyError(f"{from_key} does not exist")
    adata.obsm[to_key] = adata.obsm[from_key]
    del adata.obsm[from_key]
    _restore_obsm_key(adata, from_key)



"""
scanpy pca
"""

import logging
import scanpy as sc
from ._obj_utils import write_embedding


def pca(adata, key_added=None, remove_genes=None, export_embedding=None, **kwargs):
    """
    Wrapper function for sc.pp.pca, for supporting named slot
    """
    if isinstance(remove_genes, (tuple, list)):
        pass
    elif isinstance(remove_genes, str):
        remove_genes = [remove_genes]
    else:
        remove_genes = []
    if remove_genes:
        adata.var["hvg_full"] = adata.var["highly_variable"].copy()
        for g in remove_genes:
            if g in adata.var.columns and (
                adata.var[g].dtype.kind == "b" or isinstance(adata.var[g][0], bool)
            ):
                adata.var["highly_variable"] = adata.var["highly_variable"] & ~adata.var[g]

    # n_comps may be greater than the number of cells (n_obs), which will
    # produce an error. Additional logic may be required in future for very
    # small gene numbers (adata.n_vars).

    if "n_comps" in kwargs and kwargs["n_comps"] is not None:
        if kwargs["n_comps"] > adata.n_obs:
            logging.warning("n_comps exceeds cell number, resetting to %d", adata.n_obs)
            kwargs["n_comps"] = adata.n_obs

    # omit "svd_solver" to let scanpy choose automatically
    if "svd_solver" in kwargs and kwargs["svd_solver"] == "auto":
        del kwargs["svd_solver"]

    if key_added:
        if "X_pca" in adata.obsm.keys():
            adata.obsm["X_pca_bkup"] = adata.obsm["X_pca"]
        sc.pp.pca(adata, **kwargs)
        pca_key = f"X_pca_{key_added}"
        adata.obsm[pca_key] = adata.obsm["X_pca"]
        del adata.obsm["X_pca"]
        if "X_pca_bkup" in adata.obsm.keys():
            adata.obsm["X_pca"] = adata.obsm["X_pca_bkup"]
            del adata.obsm["X_pca_bkup"]
    else:
        sc.pp.pca(adata, **kwargs)
        pca_key = "X_pca"

    if export_embedding is not None:
        write_embedding(adata, pca_key, export_embedding, key_added=key_added)



"""
High-level pipeline functions
"""

from typing import Union, Optional

from packaging import version
import re
import numpy as np
import scipy.sparse as sp
import pandas as pd
import anndata
import scanpy as sc
from scipy.stats import norm
from sklearn.mixture import GaussianMixture
import matplotlib.pyplot as plt

from ._hvg import hvg
from ._pca import pca
from ._neighbors import neighbors
from ._umap import umap
from ._fdg import fdg
from ._tsne import tsne
from ._leiden import leiden
from ._diffmap import diffmap
from ._plot import (
    clear_colors,
    plot_qc_violin,
    plot_qc_scatter,
)
from ._annot import (
    LR_train,
    LR_predict,
)
from ._utils import (
    restore_adata,
    sc_warn,
    cross_table,
)
from ._markers import calc_marker_stats, filter_marker_stats


class QcLowPassError(ValueError):
    """
    Error raised when a QC minimum requirements are not met for a dataset.
    """

    pass


def calculate_qc(
    adata,
    flags={"mito": r"^MT-", "ribo": r"^RP[LS]", "hb": r"^HB"},
    extra_flags={},
    flag_only=False,
    suffix="",
    **kwargs,
) -> None:
    """
    Calculate quality control (QC) metrics for an AnnData object. The object is
    modified in-place.

    Args:
        adata: AnnData object to calculate QC metrics for.
        flags: Dictionary of QC flags and regular expression patterns to match
            gene names against.
        extra_flags: Additional QC flags and patterns to add to `flags`.
        flag_only: If True, only calculate QC flags and do not calculate other
            metrics.
        suffix: Suffix to append to QC metric names.
        **kwargs: Additional keyword arguments to pass to
            :py:func:`scanpy.pp.calculate_qc_metrics`.

    Returns:
        None.

    Raises:
        None.

    Examples:
        >>> import scanpy as sc
        >>> import sctk
        >>> adata = sc.datasets.pbmc3k()
        >>> sctk.calculate_qc(adata)

    """
    if extra_flags:
        for k, v in extra_flags.items():
            flags[k] = v

    qc_vars = []
    for flag, pattern in flags.items():
        if flag not in adata.var.columns and pattern:
            pat = re.compile(pattern)
            adata.var[flag] = np.array(
                [bool(pat.search(g)) for g in adata.var_names]
            )
        if flag in adata.var.columns:
            qc_vars.append(flag)

    if flag_only:
        return

    qc_vars.extend([flag for flag in extra_flags if flag in adata.var.columns])

    qc_tbls = sc.pp.calculate_qc_metrics(
        adata, qc_vars=qc_vars, percent_top=[50], **kwargs
    )

    adata.obs[f"n_counts{suffix}"] = qc_tbls[0]["total_counts"].values
    adata.obs[f"log1p_n_counts{suffix}"] = np.log1p(
        adata.obs[f"n_counts{suffix}"]
    )
    adata.obs[f"n_genes{suffix}"] = qc_tbls[0]["n_genes_by_counts"].values
    adata.obs[f"log1p_n_genes{suffix}"] = np.log1p(
        adata.obs[f"n_genes{suffix}"]
    )
    for metric in qc_vars:
        adata.obs[f"percent_{metric}{suffix}"] = qc_tbls[0][
            f"pct_counts_{metric}"
        ].values
        adata.obs[f"n_counts_{metric}{suffix}"] = (
            qc_tbls[0][f"pct_counts_{metric}"].values
            * qc_tbls[0]["total_counts"].values
            / 100
        )
    adata.obs[f"percent_top50{suffix}"] = qc_tbls[0][
        "pct_counts_in_top_50_genes"
    ].values
    adata.var[f"n_counts{suffix}"] = qc_tbls[1]["total_counts"].values
    adata.var[f"n_cells{suffix}"] = qc_tbls[1]["n_cells_by_counts"].values


def generate_qc_clusters(
    ad,
    metrics,
    aux_ad=None,
    n_pcs=None,
    n_neighbors=None,
    res=0.2,
    return_aux=False,
) -> anndata.AnnData:
    """
    Generate quality control (QC) clusters for an AnnData object. The object is
    modified in-place.

    This function generates QC clusters for an AnnData object using the
    specified QC metrics. If an auxiliary AnnData object is not provided, this
    function will create one by performing PCA, nearest neighbor graph
    construction, and UMAP embedding on the specified QC metrics.

    Args:
        ad: AnnData object to generate QC clusters for.
        metrics: List of QC metrics to use for generating QC clusters. Must be
            present as obs columns.
        aux_ad: Optional auxiliary AnnData object to use for generating QC
            clusters, created by an earlier call of this function and returned if
            return_aux is set to True. Its neighbour graph will be used for
            clustering and its UMAP will be transferred to the input object.
        n_pcs: Number of principal components to use for PCA. If not provided,
            this will be set to max(2, len(metrics) - 2).
        n_neighbors: Number of nearest neighbors to use for constructing the
            nearest neighbor graph. If not provided, this will be set to min(max(5,
            int(ad.n_obs / 500)), 10).
        res: Resolution parameter to use for the Leiden clustering algorithm.
        return_aux: If True, return the auxiliary AnnData object used for
            generating QC clusters.

    Returns:
        If `return_aux` is False, returns None. Otherwise, returns the auxiliary
        AnnData object used for generating QC clusters.

    Raises:
        None.

    Examples:
        >>> import scanpy as sc
        >>> import sctk
        >>> adata = sc.datasets.pbmc3k()
        >>> sctk.calculate_qc(adata)
        >>> metrics_list = ["n_counts", "n_genes", "percent_mito", "percent_ribo", "percent_hb"]
        >>> sctk.generate_qc_clusters(adata, metrics=metrics_list)

    """
    if aux_ad is None:
        n_pcs = max(2, len(metrics) - 2) if n_pcs is None else n_pcs
        n_neighbors = min(max(5, int(ad.n_obs / 500)), 10)
        aux_ad = anndata.AnnData(
            ad.obs[metrics].values,
            obs=ad.obs.copy(),
            var=pd.DataFrame(index=["var_" + m for m in metrics]),
        )
        sc.pp.scale(aux_ad)
        sc.pp.pca(aux_ad, n_comps=n_pcs)
        sc.pp.neighbors(aux_ad, n_neighbors=n_neighbors)
        sc.tl.umap(aux_ad, min_dist=0.1)
    sc.tl.leiden(aux_ad, resolution=res, key_added="qc_cluster")
    ad.obs["qc_cluster"] = aux_ad.obs.qc_cluster
    ad.obsm["X_umap_qc"] = aux_ad.obsm["X_umap"]
    if return_aux:
        return aux_ad


def _scale_factor(x):
    """
    Calculate the scaling factor for a given array.

    This function calculates the scaling factor for a given array by computing
    the difference between the maximum and minimum values of the array and
    scaling it to a fixed value of 5.0.

    TODO why?

    Args:
        x: Numpy array to calculate the scaling factor for.

    Returns:
        Scaling factor as a float.

    Raises:
        None.

    Examples:
        >>> import numpy as np
        >>> from sctk._pipeline import _scale_factor
        >>> x = np.array([1, 2, 3, 4, 5])
        >>> _scale_factor(x)
        1.25

    """
    xmin = np.min(x)
    xmax = np.max(x)
    return 5.0 / (xmax - xmin)


def fit_gaussian(
    x,
    n_components=np.arange(10) + 1,
    threshold=0.05,
    xmin=None,
    xmax=None,
    plot=False,
    nbins=500,
    hist_bins=100,
) -> tuple:
    """
    Fit a Gaussian mixture model to a 1D numpy array.

    This function fits a Gaussian mixture model to a 1D numpy array using the
    specified number of components and threshold. It returns the lower and upper
    bounds of the fitted Gaussian distribution, as well as the fitted Gaussian
    mixture model.

    Args:
        x: 1D numpy array to fit a Gaussian mixture model to.
        n_components: Number of components to use for the Gaussian mixture model.
            The best GMM will be selected based on BIC.
        threshold: Threshold value for determining the lower and upper bounds of
            the fitted Gaussian distribution.
        xmin: Minimum value to use for the fitted Gaussian distribution. If not
            provided, this will be set to the minimum value of `x`.
        xmax: Maximum value to use for the fitted Gaussian distribution. If not
            provided, this will be set to the maximum value of `x`.
        plot: If True, plot the fitted Gaussian distribution.
        nbins: Number of bins to use for the distribution of `x`.
        hist_bins: Number of bins to use for the histogram of `x` in the plot.

    Returns:
        Tuple containing the lower bound, upper bound, and Gaussian mixture
        model.

    Raises:
        None.

    Examples:
        >>> import numpy as np
        >>> from sctk import fit_gaussian
        >>> x = np.random.normal(loc=0, scale=1, size=1000)
        >>> x_left, x_right, gmm = fit_gaussian(x, n_components=[2], threshold=0.1)

    """
    xmin = x.min() if xmin is None else xmin
    xmax = x.max() if xmax is None else xmax
    x_fit = x[(x >= xmin) & (x <= xmax)]
    f = _scale_factor(x_fit)
    x_fit = (x_fit * f).reshape(-1, 1)
    # try a bunch of different component counts for the GMM
    gmms = []
    bics = []
    for n in n_components:
        gmm = GaussianMixture(n_components=n, random_state=0)
        gmm.fit(x_fit)
        while not gmm.converged_:
            gmm.fit(x_fit, warm_start=True)
        gmms.append(gmm)
        bics.append(gmm.bic(x_fit))
    # pick best one based on BIC (the lower the better)
    # making this plot is useless if there's a single component count
    if plot and len(n_components) > 1:
        plt.plot(n_components, bics)
        plt.xlabel("GMM components")
        plt.ylabel("BIC")
        plt.show()
    # the minimum bic's index is the position in n_components
    # as well as the gmm list
    n = n_components[np.argmin(bics)]
    gmm = gmms[np.argmin(bics)]
    x0 = np.linspace(x.min(), x.max(), num=nbins)
    y_pdf = np.zeros((n, nbins))
    y_cdf = np.zeros((n, nbins))
    for i in range(n):
        y_pdf[i] = (
            norm.pdf(
                x0 * f,
                loc=gmm.means_[i, 0],
                scale=n * gmm.covariances_[i, 0, 0],
            )
            * gmm.weights_[i]
        )
        y_cdf[i] = (
            norm.cdf(
                x0 * f,
                loc=gmm.means_[i, 0],
                scale=n * gmm.covariances_[i, 0, 0],
            )
            * gmm.weights_[i]
        )
    y0 = y_pdf.sum(axis=0)
    x_peak = x0[np.argmax(y0)]
    try:
        x_left = x0[(y0 < threshold) & (x0 < x_peak)].max()
    except Exception:
        sc_warn("Failed to find lower bound, using min value instead.")
        x_left = x0.min()
    try:
        x_right = x0[(y0 < threshold) & (x0 > x_peak)].min()
    except Exception:
        sc_warn("Failed to find upper bound, using max value instead.")
        x_right = x0.max()
    if plot:
        fig, ax1 = plt.subplots()
        _ = ax1.hist(x, bins=hist_bins)
        ax2 = ax1.twinx()
        ax2.plot(x0, y0, c="k")
        ax2.hlines(
            y=threshold,
            xmin=x.min(),
            xmax=x.max(),
            linewidth=1,
            linestyle="dotted",
        )
        ax2.vlines(
            x=[xmin, xmax],
            ymin=y0.min(),
            ymax=y0.max(),
            linewidth=1,
            linestyle="dashed",
        )
        if not np.isnan(x_left):
            ax2.vlines(x=x_left, ymin=y0.min(), ymax=y0.max(), linewidth=1)
        if not np.isnan(x_right):
            ax2.vlines(x=x_right, ymin=y0.min(), ymax=y0.max(), linewidth=1)
    return x_left, x_right, gmm


def cellwise_qc(adata, metrics=None, cell_qc_key="cell_passed_qc", **kwargs):
    """
    Filter cells in an AnnData object based on quality control metrics. The
    object is modified in-place.

    This function filters cells in an AnnData object based on quality control
    metrics. The metrics used for filtering can be specified using the `metrics`
    argument. By default, the function uses a set of default metrics, but these
    can be overridden by passing a list/tuple of metric names or a dictionary of
    metric names and their corresponding parameters.

    Args:
        adata: AnnData object to filter cells from.
        metrics: Optional list/tuple of metric names or dictionary of metric
            names and their corresponding parameters. If not provided, the function
            uses a set of default metrics. For defaults and an explanation, please
            refer to the QC workflow demo notebook.
        cell_qc_key: Obs column in the object to store the per-cell QC calls in.
        **kwargs: Additional keyword arguments to pass to the
            :py:func:`fit_gaussian` function.

    Returns:
        None.

    Raises:
        ValueError: If `metrics` is not a list/tuple of metric names or a
            dictionary of metric names and their corresponding parameters.

    Examples:
        >>> import scanpy as sc
        >>> import sctk
        >>> adata = sc.datasets.pbmc3k()
        >>> sctk.calculate_qc(adata)
        >>> sctk.cellwise_qc(adata)
    """
    default_metric_params = {
        "n_counts": (1000, None, "log", "min_only", 0.1),
        "n_genes": (100, None, "log", "min_only", 0.1),
        "percent_mito": (0.01, 20, "log", "max_only", 0.1),
        "percent_ribo": (0, 100, "log", "both", 0.1),
        "percent_hb": (None, 1, "log", "max_only", 0.1),
        "percent_soup": (None, 5, "log", "max_only", 0.1),
        "percent_spliced": (50, 97.5, "log", "both", 0.1),
        "scrublet_score": (None, 0.3, "linear", "max_only", 0.95),
    }
    if metrics is None:
        metric_params = default_metric_params
    elif isinstance(metrics, (list, tuple)):
        metric_params = {
            k: v for k, v in default_metric_params.items() if k in metrics
        }
    elif isinstance(metrics, dict) and all(
        k in adata.obs.columns and len(v) == 5 for k, v in metrics.items()
    ):
        metric_params = metrics
    else:
        raise ValueError(
            "`metrics` must be a list/tuple of metric names or a dict of"
            " <name>: [<min>, <max>, <scale>, <sidedness>, <min_pass_rate>]"
        )

    n_obs = adata.n_obs

    pass_filter = {}
    for m, v in metric_params.items():
        min_x, max_x, scale, side, min_pass_rate = v
        if m not in adata.obs.columns:
            continue
        x = adata.obs[m].values.astype(np.float32)
        if scale == "log":
            x = np.log1p(x)
            min_x = np.log1p(min_x) if min_x is not None else None
            max_x = np.log1p(max_x) if max_x is not None else None
        try:
            x_low, x_high, _ = fit_gaussian(x, xmin=min_x, xmax=max_x, **kwargs)
        except ValueError:
            x_low = min_x if min_x is not None else x.min()
            x_high = max_x if max_x is not None else x.max()
        if side == "min_only":
            x_high = x.max()
        elif side == "max_only":
            x_low = x.min()
        else:
            pass
        m_pass = (x_low <= x) & (x <= x_high)
        if m_pass.sum() < n_obs * min_pass_rate:
            if side == "min_only":
                x_low = min_x
                m_pass = min_x <= x
            elif side == "max_only":
                x_high = max_x
                m_pass = x <= max_x
            else:
                x_low, x_high = min_x, max_x
                m_pass = (min_x <= x) & (x <= max_x)
        pass_filter[m] = m_pass

        x_low_str = x_low if scale == "linear" else np.expm1(x_low)
        x_high_str = x_high if scale == "linear" else np.expm1(x_high)
        print(
            f"{m}: [{x_low_str}, {x_high_str}], {pass_filter[m].sum()}/{n_obs} passed"
        )

    all_passed = np.ones(n_obs).astype(bool)
    for m, k_pass in pass_filter.items():
        all_passed = all_passed & k_pass
    print(f"{all_passed.sum()}/{n_obs} pass")
    adata.obs[cell_qc_key] = all_passed
    if adata.obs[cell_qc_key].sum() == 0:
        print(
            "No cells passed. Performing simple filtering on counts, genes and mito%"
        )
        adata.obs[cell_qc_key] = (
            (adata.obs.n_counts >= metrics["n_counts"][0])
            & (adata.obs.n_genes >= metrics["n_genes"][0])
            & (adata.obs.percent_mito < metrics["percent_mito"][1])
        )


def filter_qc_outlier_legacy(
    adata,
    metrics=[
        "n_counts",
        "n_genes",
        "percent_mito",
        "percent_ribo",
        "percent_hb",
        "percent_top50",
    ],
    min_count=1000,
    min_gene=100,
    min_mito=0.01,
    max_mito=20,
    min_ribo=0,
    max_ribo=100,
    max_hb=1.0,
    min_top50=0,
    max_top50=100,
    min_pass_rate=0.6,
    onesided=False,
    force=False,
) -> np.ndarray:
    """
    Filter out cells with outlier QC metrics.

    This function filters out cells with outlier QC metrics based on a set of
    predefined thresholds. The function takes an AnnData object as input and
    returns a boolean array indicating which cells pass the QC filter.

    Args:
        adata: AnnData object to filter.
        metrics: List of QC metrics to use for filtering.
        min_count: Minimum number of counts per cell.
        min_gene: Minimum number of genes per cell.
        min_mito: Minimum percentage of mitochondrial genes per cell.
        max_mito: Maximum percentage of mitochondrial genes per cell.
        min_ribo: Minimum percentage of ribosomal genes per cell.
        max_ribo: Maximum percentage of ribosomal genes per cell.
        max_hb: Maximum percentage of hemoglobin genes per cell.
        min_top50: Minimum percentage of reads mapping to the top 50 genes.
        max_top50: Maximum percentage of reads mapping to the top 50 genes.
        min_pass_rate: Minimum pass rate for the QC filter.
        onesided: Whether to use a one-sided or two-sided threshold for count
            and gene metrics.
        force: Whether to force the function to pass even if the pass rate is
            below the minimum threshold.

    Returns:
        Boolean array indicating which cells pass the QC filter.

    Raises:
        QcLowPassError: If the pass rate for a given metric is below the minimum
            threshold and force is False.

    Examples: TODO check
        >>> import scanpy as sc
        >>> adata = sc.datasets.pbmc3k()
        >>> adata = adata[:, adata.var["highly_variable"]]
        >>> adata = sc.pp.normalize_total(adata, target_sum=1e4)
        >>> adata = sc.pp.log1p(adata)
        >>> adata = sc.pp.highly_variable_genes(adata, min_mean=0.0125, max_mean=3, min_disp=0.5)
        >>> adata = adata[:, adata.var["highly_variable"]]
        >>> sc.pp.pca(adata, n_comps=50, use_highly_variable=True)
        >>> sc.pp.neighbors(adata, n_neighbors=10, n_pcs=50)
        >>> sc.tl.umap(adata)
        >>> k_pass = filter_qc_outlier_legacy(adata)
    """
    k_pass = np.ones(adata.n_obs).astype(bool)

    if "n_counts" in metrics:
        try:
            x_low, x_high, _ = fit_gaussian(
                adata.obs["log1p_n_counts"].values, xmin=np.log1p(min_count)
            )
        except ValueError:
            x_low, x_high = (
                np.log1p(min_count),
                adata.obs["log1p_n_counts"].max(),
            )
        x_low = max(np.log1p(250), x_low)
        min_count = int(np.expm1(x_low))
        max_count = (
            adata.obs["n_counts"].max() if onesided else int(np.expm1(x_high))
        )
        k_count = (adata.obs["n_counts"] >= min_count) & (
            adata.obs["n_counts"] <= max_count
        )
        print(
            f"n_counts: [{min_count}, {max_count}], {k_count.sum()}/{adata.n_obs} pass"
        )
        if k_count.sum() < adata.n_obs * min_pass_rate and not force:
            raise QcLowPassError("n_counts")
        k_pass = k_pass & k_count

    if "n_genes" in metrics:
        try:
            x_low, x_high, _ = fit_gaussian(
                adata.obs["log1p_n_genes"].values, xmin=np.log1p(min_gene)
            )
        except ValueError:
            x_low, x_high = np.log1p(min_gene), adata.obs["log1p_n_genes"].max()
        x_low = max(np.log1p(50), x_low)
        min_gene = int(np.expm1(x_low))
        max_gene = (
            adata.obs["n_genes"].max() if onesided else int(np.expm1(x_high))
        )
        k_gene = (adata.obs["n_genes"] >= min_gene) & (
            adata.obs["n_genes"] <= max_gene
        )
        print(
            f"n_genes: [{min_gene}, {max_gene}], {k_gene.sum()}/{adata.n_obs} pass"
        )
        if k_gene.sum() < adata.n_obs * min_pass_rate and not force:
            raise QcLowPassError("n_genes")
        k_pass = k_pass & k_gene

    if "percent_mito" in metrics:
        if (adata.obs["percent_mito"].values > 0).sum() > 0:
            x_low, x_high, _ = fit_gaussian(
                np.log1p(adata.obs["percent_mito"].values),
                xmin=np.log1p(min_mito),
                xmax=np.log1p(max_mito),
            )
            max_mito = np.expm1(x_high)
        k_mito = adata.obs["percent_mito"] <= max_mito
        print(
            f"percent_mito: [0, {max_mito}], {k_mito.sum()}/{adata.n_obs} pass"
        )
        if k_mito.sum() < adata.n_obs * min_pass_rate and not force:
            raise QcLowPassError("percent_mito")
        k_pass = k_pass & k_mito

    if "percent_ribo" in metrics:
        x_low, x_high, _ = fit_gaussian(
            np.log1p(adata.obs["percent_ribo"].values),
            xmin=np.log1p(min_ribo),
            xmax=np.log1p(max_ribo),
        )
        min_ribo = np.expm1(x_low)
        max_ribo = np.expm1(x_high)
        k_ribo = (adata.obs["percent_ribo"] >= min_ribo) & (
            adata.obs["percent_ribo"] <= max_ribo
        )
        print(
            f"percent_ribo: [{min_ribo}, {max_ribo}], {k_ribo.sum()}/{adata.n_obs} pass"
        )
        if k_ribo.sum() < adata.n_obs * min_pass_rate and not force:
            raise QcLowPassError("percent_ribo")
        k_pass = k_pass & k_ribo

    if "percent_hb" in metrics:
        k_hb = adata.obs["percent_hb"] <= max_hb
        print(f"percent_hb: [0, {max_hb}], {k_hb.sum()}/{adata.n_obs} pass")
        k_pass = k_pass & k_hb
        if k_hb.sum() < adata.n_obs * min_pass_rate and not force:
            raise QcLowPassError("percent_hb")

    if "percent_top50" in metrics:
        x_low, x_high, _ = fit_gaussian(
            adata.obs["percent_top50"].values, xmin=min_top50, xmax=max_top50
        )
        max_top50 = x_high
        min_top50 = x_low
        k_top50 = (adata.obs["percent_top50"] <= max_top50) & (
            adata.obs["percent_top50"] >= min_top50
        )
        print(
            f"percent_top50: [{min_top50}, {max_top50}], {k_top50.sum()}/{adata.n_obs} pass"
        )
        if k_top50.sum() < adata.n_obs * min_pass_rate and not force:
            raise QcLowPassError("percent_top50")
        k_pass = k_pass & k_top50

    if "scrublet_score_z" in metrics:
        x_low, x_high, _ = fit_gaussian(
            adata.obs["scrublet_score_z"].values,
            n_components=[10],
            xmin=-99,
            xmax=10,
            threshold=0.05,
        )
        max_sz = x_high
        min_sz = x_low
        k_sz = adata.obs["scrublet_score_z"] <= max_sz
        print(
            f"scrublet_score_z: [{min_sz}, {max_sz}], {k_sz.sum()}/{adata.n_obs} pass"
        )
        if k_sz.sum() < adata.n_obs * min_pass_rate and not force:
            raise QcLowPassError("scrublet_score_z")
        k_pass = k_pass & k_sz

    print(f"{k_pass.sum()}/{adata.n_obs} pass")
    return k_pass


def clusterwise_qc(
    ad,
    threshold=0.5,
    cell_qc_key="cell_passed_qc",
    key_added="cluster_passed_qc",
) -> None:
    """
    Find good quality control (QC) clusters in an AnnData object.

    This function finds good quality control (QC) clusters in an AnnData object
    by identifying clusters that have a high proportion of cells that pass the
    QC filter.

    Args:
        ad: AnnData object to find good QC clusters in. Needs qc_cluster
            present in obs.
        threshold: Clusters featuring at least this fraction of good QC cells
            will be deemed good QC clusters.
        cell_qc_key: Key to use to retrieve per-cell QC calls from obs in the
            AnnData.
        key_added: Key to use for storing the results in the AnnData obs object.

    Returns:
        None.

    Raises:
        None.

    Examples:
        >>> import scanpy as sc
        >>> import sctk
        >>> adata = sc.datasets.pbmc3k()
        >>> sctk.calculate_qc(adata)
        >>> metrics_list = ["n_counts", "n_genes", "percent_mito", "percent_ribo", "percent_hb"]
        >>> sctk.generate_qc_clusters(adata, metrics=metrics_list)
        >>> sctk.cellwise_qc(adata)
        >>> sctk.clusterwise_wc(adata)

    """
    if ad.obs[cell_qc_key].astype(bool).sum() == 0:
        good_qc_clusters = []
    else:
        good_qc_clusters = (
            pd.crosstab(
                ad.obs.qc_cluster,
                ad.obs[cell_qc_key].astype("category"),
                normalize="index",
            )
            .where(lambda x: x[True] >= threshold)
            .dropna()
            .index.tolist()
        )

    ad.obs[key_added] = ad.obs["qc_cluster"].isin(good_qc_clusters)


def get_good_sized_batch(batches, min_size=10) -> list:
    """
    Get a list of batches with a minimum size.

    This function takes a pandas Series of batch IDs and their corresponding
    sizes and returns a list of batch IDs that have a size greater than or equal
    to the specified minimum size.

    Args:
        batches: Pandas Series of batch IDs and their corresponding sizes.
        min_size: Minimum size of batches to include in the output list. Default
            is 10.

    Returns:
        List of batch IDs with a size greater than or equal to `min_size`.

    Raises:
        None.
    """
    x = batches.value_counts()
    return x.index[x >= min_size].to_list()


def simple_default_pipeline(
    adata: anndata.AnnData,
    filter_only: bool = False,
    post_filter_only: bool = False,
    norm_only: bool = False,
    post_norm_only: bool = False,
    hvg_only: bool = False,
    post_hvg_only: bool = False,
    pca_only: bool = False,
    post_pca_only: bool = False,
    do_clustering: bool = False,
    zero_center: bool = None,
    do_combat: bool = False,
    batch: Union[str, Union[list, tuple], None] = None,
    batch_method: Optional[str] = "harmony",
    random_state: int = 0,
    clustering_resolution: Union[float, list, tuple] = [
        0.1,
        0.3,
        0.5,
        0.7,
        0.9,
    ],
    use_gpu: Union[bool, str] = False,
    use_hvg: Union[list, tuple, None] = None,
    filter_kw: dict = {},
    hvg_kw: dict = {},
    rgs_kw: dict = {},
    pca_kw: dict = {},
    nb_kw: dict = {},
    umap_kw: dict = {},
    hm_kw: dict = {},
    bk_kw: dict = {},
) -> anndata.AnnData:
    """
    Applies a default preprocessing pipeline to the input AnnData object.

    Args:
        adata: AnnData object to preprocess.
        filter_only: If True, only performs filtering steps and returns the
            filtered AnnData object.
        post_filter_only: If True, only performs post-filtering steps and
            returns the processed AnnData object.
        norm_only: If True, only performs normalization steps and returns the
            normalized AnnData object.
        post_norm_only: If True, only performs post-normalization steps and
            returns the processed AnnData object.
        hvg_only: If True, only performs highly variable gene selection and
            returns the selected AnnData object.
        post_hvg_only: If True, only performs post-highly variable gene
            selection steps and returns the processed AnnData object.
        pca_only: If True, only performs PCA and returns the PCA-transformed
            AnnData object.
        post_pca_only: If True, only performs post-PCA steps and returns the
            processed AnnData object.
        do_clustering: If True, performs clustering and updates the AnnData
            object with the clustering results.
        zero_center: If True, centers the data to zero mean before scaling. If
            None, uses a heuristic based on the number of cells.
        do_combat: If True and `batch` is not None, performs batch correction
            using ComBat.
        batch: The batch variable(s) to use for batch correction. If None, no
            batch correction is performed.
        batch_method: The batch correction method to use. Can be "harmony" or
            "bbknn".
        random_state: The random seed to use for reproducibility.
        clustering_resolution: The clustering resolution(s) to use for Leiden
            clustering.
        use_gpu: If True, uses the GPU for some computations. If "all", uses the
            GPU for all computations.
        use_hvg: The list of highly variable genes to use for normalization and
            PCA. If None, uses the default Seurat v3 HVG selection.
        filter_kw: Additional keyword arguments to pass to the
            `filter_qc_outlier_legacy` function.
        hvg_kw: Additional keyword arguments to pass to the `hvg` function.
        rgs_kw: Additional keyword arguments to pass to the `sc.pp.regress_out`
            function.
        pca_kw: Additional keyword arguments to pass to the `pca` function.
        nb_kw: Additional keyword arguments to pass to the `sc.pp.neighbors`
            function.
        umap_kw: Additional keyword arguments to pass to the `umap` function.
        hm_kw: Additional keyword arguments to pass to the `run_harmony`
            function.
        bk_kw: Additional keyword arguments to pass to the `run_bbknn` function.

    Returns:
        The preprocessed AnnData object.
    """
    if version.parse(sc.__version__) < version.parse("1.6"):
        use_gpu = False
    if not (post_filter_only or post_norm_only or post_pca_only):
        if not np.all(
            pd.Series(
                [
                    "n_counts",
                    "n_genes",
                    "percent_mito",
                    "percent_ribo",
                    "percent_hb",
                    "percent_top50",
                ]
            ).isin(list(adata.obs.columns))
        ):
            calculate_qc(adata)
        if (adata.obs["n_counts"] == 0).sum() > 0:
            adata = adata[adata.obs["n_counts"] > 0].copy()
        k_cell = filter_qc_outlier_legacy(adata, **filter_kw)
        if batch:
            if isinstance(batch, (list, tuple)):
                for b in batch:
                    if b in adata.obs.columns:
                        batches = get_good_sized_batch(adata.obs.loc[k_cell, b])
                        k_cell = k_cell & adata.obs[b].isin(batches)
            elif isinstance(batch, str):
                batches = get_good_sized_batch(adata.obs.loc[k_cell, batch])
                k_cell = k_cell & adata.obs[batch].isin(batches)
            else:
                raise ValueError("Invalid type of `batch`")
        adata = adata[k_cell, :].copy()
        if filter_only:
            return adata

    subset_hvg = hvg_kw.get("subset", False)
    if not (post_norm_only or post_pca_only):
        if not use_hvg and hvg_kw.get("flavor", "seurat") == "seurat_v3":
            aux_ad = anndata.AnnData(
                X=adata.X, obs=adata.obs.copy(), var=adata.var.copy()
            )
            if not sc._utils.check_nonnegative_integers(aux_ad.X):
                aux_ad.X = np.round(aux_ad.X)
                n_cells = (adata.layers["counts"] > 0).sum(axis=0)
            n_cells = n_cells.A1 if sp.issparse(aux_ad.X) else n_cells
            k_gene = n_cells >= 10
            if (~k_gene).sum() > 0:
                aux_ad = aux_ad[:, k_gene].copy()
            hvg_kw["subset"] = False
            hvg(aux_ad, **hvg_kw)
            use_hvg = aux_ad.var_names[aux_ad.var.highly_variable].to_list()
            adata.uns["hvg"] = aux_ad.uns["hvg"]
            del aux_ad
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
        if norm_only:
            return adata

    if not post_pca_only:
        if adata.raw is None:
            adata.raw = adata
        else:
            adata.X = adata.raw.X[
                :, adata.raw.var_names.isin(adata.var_names)
            ].copy()
        if "n_counts" not in adata.var.keys():
            n_counts = np.expm1(adata.X).sum(axis=0)
            adata.var["n_counts"] = (
                n_counts.A1 if sp.issparse(adata.X) else n_counts
            )
        n_cells = (adata.X > 0).sum(axis=0)
        adata.var["n_cells"] = n_cells.A1 if sp.issparse(adata.X) else n_cells
        k_gene = adata.var["n_cells"] >= 3
        if (~k_gene).sum() > 0:
            adata = adata[:, k_gene].copy()
        if batch and do_combat:
            sc.pp.combat(adata, key=batch)
        if use_hvg:
            adata.var["highly_variable"] = adata.var_names.isin(use_hvg)
            if subset_hvg:
                adata = adata[adata.var["highly_variable"]].copy()
        elif hvg_kw.get("flavor", "seurat") != "seurat_v3":
            hvg(adata, **hvg_kw)
        if hvg_only:
            return adata
        zero_center = (
            (adata.n_obs <= 20000) if zero_center is None else zero_center
        )
        if zero_center or rgs_kw:
            adata = adata[:, adata.var.highly_variable.values].copy()
        if rgs_kw:
            sc.pp.regress_out(adata, **rgs_kw)
        sc.pp.scale(adata, zero_center=zero_center, max_value=10)
        n_comps = min(30, adata.n_obs - 1, adata.var.highly_variable.sum() - 1)
        if n_comps < 2:
            raise ValueError("n_obs or n_vars too small for pca calculation")
        pca(
            adata,
            n_comps=n_comps,
            zero_center=zero_center,
            svd_solver="arpack",
            use_highly_variable=True,
            **pca_kw,
        )
        if pca_only:
            return adata

    n_neighbors = nb_kw.pop("n_neighbors", 15)
    n_pcs = nb_kw.pop("n_pcs", 20)
    if batch:
        if batch_method == "bbknn":
            from ._utils import run_bbknn

            key_added = bk_kw.pop("key_added", "bk")
            if use_gpu == "all":
                bk_kw["metric"] = "euclidean"
            run_bbknn(adata, batch=batch, key_added=key_added, **bk_kw)
            umap(
                adata,
                method="rapids" if use_gpu == "all" else None,
                use_graph=f"neighbors_{key_added}",
                key_added=key_added,
                random_state=random_state,
                **umap_kw,
            )
            if do_clustering:
                leiden(
                    adata,
                    flavor="rapids" if use_gpu else None,
                    use_graph=f"neighbors_{key_added}",
                    resolution=clustering_resolution,
                    key_added=key_added,
                )
        else:
            from ._utils import run_harmony

            key_added = hm_kw.pop("key_added", "hm")
            run_harmony(adata, batch=batch, key_added=key_added, **hm_kw)
            neighbors(
                adata,
                method="rapids" if use_gpu else "umap",
                use_rep=f"X_pca_{key_added}",
                key_added=key_added,
                n_pcs=n_pcs,
                n_neighbors=n_neighbors,
            )
            umap(
                adata,
                method="rapids" if use_gpu == "all" else None,
                use_graph=f"neighbors_{key_added}",
                key_added=key_added,
                random_state=random_state,
                **umap_kw,
            )
            if do_clustering:
                leiden(
                    adata,
                    flavor="rapids" if use_gpu else None,
                    use_graph=f"neighbors_{key_added}",
                    resolution=clustering_resolution,
                    key_added=key_added,
                )
    else:
        sc.pp.neighbors(
            adata, use_rep="X_pca", n_pcs=n_pcs, n_neighbors=n_neighbors
        )
        umap(
            adata,
            method="rapids" if use_gpu == "all" else None,
            random_state=random_state,
            **umap_kw,
        )
        if do_clustering:
            leiden(
                adata,
                flavor="rapids" if use_gpu else None,
                resolution=clustering_resolution,
            )

    return adata


def recluster_subset(
    adata, groupby, groups, res, new_key, ad_aux=None, **kwargs
):
    """
    Reclusters a subset of cells in an AnnData object.

    This function reclusters a subset of cells in an AnnData object by first
    filtering, normalizing, and performing PCA on the subset, and then
    performing clustering using the Leiden algorithm. The resulting clusters are
    added to the original AnnData object as a new categorical variable.

    Args:
        adata: AnnData object to recluster.
        groupby: Categorical variable to group cells by.
        groups: List of group names to recluster.
        res: Clustering resolution to use for Leiden clustering.
        new_key: Name of the new categorical variable to add to the AnnData
            object.
        ad_aux: Optional preprocessed AnnData object to use for clustering.
        **kwargs: Additional keyword arguments to pass to the
            :py:func:`simple_default_pipeline` function.

    Returns:
        If `ad_aux` is not provided, the preprocessed AnnData object used for
        clustering. Otherwise, None.
    """

    kwargs["post_norm_only"] = True
    kwargs["do_clustering"] = False
    if isinstance(res, (list, tuple)):
        res = res[0]
    if "batch" in kwargs:
        graph = (
            "neighbors_bk"
            if kwargs["batch_method"] == "bbknn"
            else "neighbors_hm"
        )
    else:
        graph = "neighbors"
    k_groups = adata.obs[groupby].isin(groups)
    if ad_aux is None:
        ad = restore_adata(
            adata[k_groups].copy(), restore_type="norm", use_raw=True
        )
        ad_aux = simple_default_pipeline(ad, **kwargs)
        return_ad = True
    else:
        return_ad = False
    leiden(ad_aux, use_graph=graph, resolution=res, key_added="aux")
    adata.obs[new_key] = adata.obs[groupby].astype(str)
    adata.obs.loc[k_groups, new_key] = (
        "_".join(groups) + "," + ad_aux.obs["leiden_aux"].astype(str)
    )
    if return_ad:
        return ad_aux


def auto_zoom_in(
    ad,
    use_graph,
    groupby,
    restrict_to=None,
    start_index=1,
    min_res=0.1,
    max_res=0.7,
    leiden_kw={},
    marker_kw={"max_next_frac": 0.3},
) -> None:
    """
    Automatically zooms in on subclusters in an AnnData object.

    This function performs iterative clustering on an AnnData object using the
    Leiden algorithm, starting from a specified categorical variable and
    splitting each group into subclusters until no further subclustering is
    possible. The resulting subclusters are added to the original AnnData object
    as new categorical variables.

    Args:
        ad: AnnData object to cluster.
        use_graph: Name of the graph to use for clustering.
        groupby: Name of the categorical variable to start clustering from.
        restrict_to: Optional list of group names to restrict clustering to.
        start_index: Index to start numbering new categorical variables from.
        min_res: Minimum clustering resolution to use.
        max_res: Maximum clustering resolution to use.
        leiden_kw: Additional keyword arguments to pass to the `leiden`
            function.
        marker_kw: Additional keyword arguments to pass to the
            :py:func:`filter_marker_stats` function.

    Returns:
        None.
    """
    i = start_index
    if isinstance(restrict_to, (list, tuple)):
        groups = restrict_to
    else:
        groups = ad.obs[groupby].cat.categories
    prev_grp_key = groupby
    for grp in groups:
        res = min_res
        new_grp_key = f"leiden_split{i}"
        subclustered = False
        print([grp, i], end=" ")
        while True:
            leiden(
                ad,
                use_graph=use_graph,
                restrict_to=(prev_grp_key, [grp]),
                resolution=res,
                key_added=f"split{i}",
                **leiden_kw,
            )
            if (
                ad.obs[new_grp_key].cat.categories.size
                > ad.obs[prev_grp_key].cat.categories.size
            ):
                subclustered = True
                break
            elif res < max_res:
                res += 0.1
            else:
                del ad.obs[new_grp_key]
                break
        if subclustered:
            print("find_marker", end=" ")
            new_groups = ad.obs[new_grp_key].cat.categories[
                ~ad.obs[new_grp_key].cat.categories.isin(
                    ad.obs[prev_grp_key].cat.categories
                )
            ]
            mkst = calc_marker_stats(ad, groupby=new_grp_key)
            mks = filter_marker_stats(mkst[2], **marker_kw)
            if (
                new_groups.isin(mks.top_frac_group.unique()).sum()
                == new_groups.size
            ):
                prev_grp_key = new_grp_key
                i += 1
                print("done", end="")
            else:
                del ad.obs[new_grp_key]
                print("failed", end="")
        print("")
    if new_grp_key in ad.obs.columns:
        print(new_grp_key)
    else:
        print(f"leiden_split{i-1}")


def custom_pipeline(
    adata,
    qc_only=False,
    plot=True,
    batch=None,
    filter_params={
        "min_genes": 200,
        "min_cells": 3,
        "max_counts": 25000,
        "max_mito": 20,
        "min_mito": 0,
    },
    norm_params={"target_sum": 1e4, "fraction": 0.9},
    combat_args={"key": None},
    hvg_params={"flavor": "seurat", "by_batch": None},
    scale_params={"max_value": 10},
    pca_params={
        "n_comps": 50,
        "svd_solver": "arpack",
        "use_highly_variable": True,
    },
    harmony_params={"batch": None, "theta": 2.0},
    nb_params={"n_neighbors": 15, "n_pcs": 20},
    umap_params={},
    tsne_params={},
    diffmap_params={"n_comps": 15},
    leiden_params={
        "resolution": [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
    },
    fdg_params={"layout": "fa"},
) -> anndata.AnnData:
    """
    Applies a custom Scanpy pipeline to an AnnData object.

    This function applies a custom Scanpy pipeline to an AnnData object, which
    includes filtering, normalization, batch correction, highly variable gene
    selection, dimensionality reduction, clustering, and visualization. The
    pipeline can be customized by passing various keyword arguments to the
    function.

    Args:
        adata: AnnData object to apply the pipeline to.
        qc_only: If True, only perform quality control and skip the rest of the
            pipeline.
        plot: If True, generate plots of the pipeline results.
        batch: Categorical variable to correct for batch effects.
        filter_params: Dictionary of filtering parameters to pass to the
            `sc.pp.filter_cells` and `sc.pp.filter_genes` functions.
        norm_params: Dictionary of normalization parameters to pass to the
            `sc.pp.normalize_total` function.
        combat_args: Dictionary of batch correction parameters to pass to the
            `sc.pp.combat` function.
        hvg_params: Dictionary of highly variable gene selection parameters to
            pass to the `hvg` function.
        scale_params: Dictionary of scaling parameters to pass to the
            `sc.pp.scale` function.
        pca_params: Dictionary of PCA parameters to pass to the `pca` function.
        harmony_params: Dictionary of Harmony parameters to pass to the
            `run_harmony` function.
        nb_params: Dictionary of nearest neighbor parameters to pass to the
            `neighbors` function.
        umap_params: Dictionary of UMAP parameters to pass to the `umap`
            function.
        tsne_params: Dictionary of t-SNE parameters to pass to the `tsne`
            function.
        diffmap_params: Dictionary of diffusion map parameters to pass to the
            `diffmap` function.
        leiden_params: Dictionary of Leiden clustering parameters to pass to the
            `leiden` function.
        fdg_params: Dictionary of force-directed graph parameters to pass to the
            `fdg` function.

    Returns:
        The modified AnnData object.

    TODO some functions return a modified anndata object, some modify in place
    """
    if qc_only:
        calculate_qc(adata)
        if plot:
            plot_qc_violin(
                adata,
                groupby=batch,
                one_per_line=adata.obs[batch].cat.categories.size > 10,
            )
            plot_qc_scatter(adata, use_hexbin=True)
    else:
        if filter_params is not None and isinstance(filter_params, dict):
            if "min_genes" in filter_params:
                sc.pp.filter_cells(adata, min_genes=filter_params["min_genes"])
            if "min_cells" in filter_params:
                sc.pp.filter_genes(adata, min_cells=filter_params["min_cells"])
            if "min_counts" in filter_params:
                k = adata.obs["n_counts"] >= filter_params["min_counts"]
                adata._inplace_subset_obs(k)
            if "max_counts" in filter_params:
                k = adata.obs["n_counts"] <= filter_params["max_counts"]
                adata._inplace_subset_obs(k)
            if "min_mito" in filter_params:
                k = adata.obs["percent_mito"] >= filter_params["min_mito"]
                adata._inplace_subset_obs(k)
            if "max_mito" in filter_params:
                k = adata.obs["percent_mito"] <= filter_params["max_mito"]
                adata._inplace_subset_obs(k)
            if "min_ribo" in filter_params:
                k = adata.obs["percent_ribo"] >= filter_params["min_ribo"]
                adata._inplace_subset_obs(k)
            if "max_ribo" in filter_params:
                k = adata.obs["percent_ribo"] <= filter_params["max_ribo"]
                adata._inplace_subset_obs(k)
            if "min_hb" in filter_params:
                k = adata.obs["percent_hb"] >= filter_params["min_hb"]
                adata._inplace_subset_obs(k)
            if "max_hb" in filter_params:
                k = adata.obs["percent_hb"] <= filter_params["max_hb"]
                adata._inplace_subset_obs(k)
            if "counts" not in adata.layers.keys():
                adata.layers["counts"] = adata.X
        if norm_params is not None and isinstance(norm_params, dict):
            if "counts" in adata.layers.keys():
                adata.X = adata.layers["counts"]
            sc.pp.normalize_total(adata, **norm_params)
            sc.pp.log1p(adata)
            adata.raw = adata
        if combat_args is not None and (
            isinstance(combat_args, dict)
            and combat_args.get("key", None)
            and combat_args["key"] in adata.obs.keys()
        ):
            adata.layers["X"] = adata.X
            adata.X = adata.raw.X
            sc.pp.combat(adata, **combat_args)
        if hvg_params is not None and isinstance(hvg_params, dict):
            hvg(adata, **hvg_params)
        if scale_params is not None and isinstance(scale_params, dict):
            sc.pp.scale(adata, **scale_params)
        if pca_params is not None and isinstance(pca_params, dict):
            pca(adata, **pca_params)
        if harmony_params is not None and (
            isinstance(harmony_params, dict)
            and harmony_params.get("batch", None)
        ):
            from ._utils import run_harmony

            run_harmony(adata, **harmony_params)
        if nb_params is not None and isinstance(nb_params, dict):
            neighbors(adata, **nb_params)
        if umap_params is not None and isinstance(umap_params, dict):
            umap(adata, **umap_params)
        if tsne_params is not None and isinstance(tsne_params, dict):
            tsne(adata, **tsne_params)
        if diffmap_params is not None and isinstance(diffmap_params, dict):
            diffmap(adata, **diffmap_params)
        if leiden_params is not None and isinstance(leiden_params, dict):
            leiden(adata, **leiden_params)
        if fdg_params is not None and isinstance(fdg_params, dict):
            fdg(adata, **fdg_params)
    return adata


def save_pipeline_object(
    ad,
    out_prefix=None,
    batch_method=None,
    obs_keys=[],
    obsm_keys=[],
    uns_keys=[],
) -> anndata.AnnData:
    """
    Saves a processed AnnData object to a file.

    This function saves a processed AnnData object to a file, after removing
    certain keys from the `obs`, `obsm`, and `uns` attributes of the object. The
    specific keys to remove depend on the `batch_method` argument, which can be
    set to "harmony", "bbknn", or None. If `batch_method` is None, the default
    keys to remove are ["leiden_r0_1", "leiden_r0_3", "leiden_r0_5",
    "leiden_r0_7", "leiden_r0_9"] and ["X_pca"]. If `batch_method` is "harmony",
    the keys to remove are ["leiden_hm_r0_1", "leiden_hm_r0_3",
    "leiden_hm_r0_5", "leiden_hm_r0_7", "leiden_hm_r0_9"], ["X_pca"], and
    ["X_pca_hm", "neighbors"]. If `batch_method` is "bbknn", the keys to remove
    are ["leiden_bk_r0_1", "leiden_bk_r0_3", "leiden_bk_r0_5", "leiden_bk_r0_7",
    "leiden_bk_r0_9"], ["X_pca"], and ["X_pca_bk", "neighbors"].

    TODO is this very custom with respect to the keys, or widely used?

    Args:
        ad: AnnData object to save.
        out_prefix: Prefix for the output file name. If None, no file is saved.
        batch_method: Batch correction method used to process the AnnData
            object. Can be "harmony", "bbknn", or None.
        obs_keys: List of keys to remove from the `obs` attribute of the AnnData
            object.
        obsm_keys: List of keys to remove from the `obsm` attribute of the
            AnnData object.
        uns_keys: List of keys to remove from the `uns` attribute of the AnnData
            object.

    Returns:
        The processed AnnData object.
    """
    if batch_method is None:
        obs_keys = [
            "leiden_r0_1",
            "leiden_r0_3",
            "leiden_r0_5",
            "leiden_r0_7",
            "leiden_r0_9",
        ]
        obsm_keys = ["X_pca"]
    elif batch_method == "harmony":
        obs_keys = [
            "leiden_hm_r0_1",
            "leiden_hm_r0_3",
            "leiden_hm_r0_5",
            "leiden_hm_r0_7",
            "leiden_hm_r0_9",
        ]
        obsm_keys = ["X_pca", "X_pca_hm"]
        uns_keys = ["neighbors"]
    elif batch_method == "bbknn":
        obs_keys = [
            "leiden_bk_r0_1",
            "leiden_bk_r0_3",
            "leiden_bk_r0_5",
            "leiden_bk_r0_7",
            "leiden_bk_r0_9",
        ]
        obsm_keys = ["X_pca", "X_pca_bk"]
        uns_keys = ["neighbors"]

    ad1 = ad.copy()
    for k in obs_keys:
        if k in ad1.obs.keys():
            del ad1.obs[k]
    for k in obsm_keys:
        if k in ad1.obsm.keys():
            del ad1.obsm[k]
    for k in uns_keys:
        if k in ad1.uns.keys():
            del ad1.uns[k]
    clear_colors(ad1)
    if ad1.raw:
        ad1.X = ad1.raw.X
        ad1.raw = None
    if out_prefix:
        ad1.write(f"{out_prefix}.processed.h5ad", compression="lzf")
    return ad1


def integrate(
    ads,
    ad_prefices=None,
    ad_types=None,
    annotations=None,
    batches=None,
    join="inner",
    n_hvg=4000,
    pool_only=False,
    normalize=True,
):
    """
    Integrates multiple AnnData objects into a single object.

    This function integrates multiple AnnData objects into a single object by
    first normalizing and preprocessing each object, and then concatenating them
    using the `concatenate` function. The resulting object can be used for
    downstream analysis, such as clustering and visualization.

    Args:
        ads: List of AnnData objects to integrate.
        ad_prefices: List of prefixes to use for the batch categories in the
            concatenated object. If not provided, uses integers starting from 0.
        ad_types: List of types to use for each AnnData object. If not provided,
            attempts to automatically determine the type of each object.
        annotations: List of annotation columns to use for each AnnData object.
            If not provided, uses the default "annot" column.
        batches: List of batch columns to use for each AnnData object. If not
            provided, uses the default "batch" column.
        join: Method to use for joining the AnnData objects. Can be "inner"
            (default) or "outer".
        n_hvg: Number of highly variable genes to select for each batch.
        pool_only: If True, returns the concatenated object without performing
            downstream analysis.
        normalize: If True, normalizes the expression data in each AnnData
            object before concatenating.

    Returns:
        If `pool_only` is True, the concatenated AnnData object. Otherwise, the
        preprocessed and analysed concatenated AnnData object.
    """
    n_ad = len(ads)
    if ad_prefices is None:
        ad_prefices = list(map(str, range(n_ad)))
    if ad_types is not None:
        if isinstance(ad_types, str):
            ad_types = [ad_types] * n_ad
        elif not isinstance(ad_types, (tuple, list)) or len(ad_types) != n_ad:
            raise ValueError("invalid `ad_types` provided")
    else:
        ad_types = ["auto"] * n_ad
    if annotations is not None:
        if isinstance(annotations, str):
            annotations = [annotations] * n_ad
        elif (
            not isinstance(annotations, (tuple, list))
            or len(annotations) != n_ad
        ):
            raise ValueError("invalid `annotations` provided")
    if batches is not None:
        if isinstance(batches, str):
            batches = [batches] * n_ad
        elif not isinstance(batches, (tuple, list)) or len(batches) != n_ad:
            raise ValueError("invalid `batches` provided")

    norm_ads = []
    for i, ad in enumerate(ads):
        ad_type = ad_types[i]
        if ad_type not in ("raw_norm", "counts", "norm"):
            if ad.raw and sp.issparse(ad.raw.X):
                ad_type = "raw_norm"
            elif (
                sp.issparse(ad.X)
                and np.abs(ad.X.data - ad.X.data.astype(int)).sum()
                < 1e-6 * ad.X.data.size
            ):
                ad_type = "counts"
            elif (
                sp.issparse(ad.X)
                and np.abs(
                    np.expm1(ad.X[0:10, :]).sum(axis=1).A1
                    - np.expm1(ad.X[0:10, :]).sum(axis=1).A1.mean()
                ).sum()
                < 1e-2
            ):
                ad_type = "norm"
            else:
                raise ValueError(
                    f"Cannot determine the type of anndata at position {i}"
                )
            print(ad_type)

        if ad_type == "raw_norm":
            norm_ad = anndata.AnnData(
                X=ad.raw.X, obs=ad.obs.copy(), var=ad.raw.var.copy()
            )
        elif ad_type == "norm":
            if not sp.issparse(ad.X):
                norm_ad = anndata.AnnData(
                    X=sp.csr_matrix(ad.X), obs=ad.obs.copy(), var=ad.var.copy()
                )
            else:
                norm_ad = anndata.AnnData(
                    X=ad.X, obs=ad.obs.copy(), var=ad.var.copy()
                )
        else:
            norm_ad = anndata.AnnData(
                X=ad.X, obs=ad.obs.copy(), var=ad.var.copy()
            )
            if normalize:
                sc.pp.normalize_total(norm_ad, target_sum=1e4)
                sc.pp.log1p(norm_ad)

        if normalize:
            post_norm_count = (
                np.expm1(norm_ad.X[0:10, :]).sum(axis=1).A1.mean().astype(int)
            )
            if post_norm_count != 10000:
                norm_ad.X = norm_ad.X / (post_norm_count / 1e4)

        prefix = ad_prefices[i]
        if (
            batches
            and batches[i] in norm_ad.obs.columns
            and batches[i] != "batch"
        ):
            if "batch" in norm_ad.obs.columns:
                del norm_ad.obs["batch"]
            norm_ad.obs.rename(columns={batches[i]: "batch"}, inplace=True)
        if (
            annotations
            and annotations[i] in norm_ad.obs.columns
            and annotations[i] != "annot"
        ):
            if "annot" in norm_ad.obs.columns:
                del norm_ad.obs["annot"]
            norm_ad.obs.rename(columns={annotations[i]: "annot"}, inplace=True)
            norm_ad.obs["annot"] = f"{prefix}_" + norm_ad.obs["annot"].astype(
                str
            )
        norm_ads.append(norm_ad)
        del norm_ad

    pooled = anndata.AnnData.concatenate(
        *norm_ads, batch_key="dataset", batch_categories=ad_prefices, join=join
    )

    calculate_qc(pooled, flag_only=True)
    if pool_only:
        return pooled
    pooled1 = simple_default_pipeline(
        pooled,
        post_norm_only=True,
        do_clustering=False,
        batch="dataset" if batches is None else ["dataset", "batch"],
        hvg_kw={"by_batch": ("dataset", 1), "n_hvg": n_hvg},
        pca_kw={"remove_genes": ("mito", "ribo")},
        hm_kw={"max_iter_harmony": 20},
    )
    return pooled1


def crossmap(adata, dataset="dataset", annotation="annot"):
    """
    Cross-maps annotations between two datasets in an AnnData object.

    This function cross-maps annotations between two datasets in an AnnData
    object by first training a logistic regression model on each dataset, and
    then using the models to predict the annotations in the other dataset. The
    resulting cross-mapped annotations are added to the original AnnData object
    as new categorical variables.

    Args:
        adata: AnnData object to cross-map annotations in.
        dataset: Name of the categorical variable to group cells by. Default is
            "dataset".
        annotation: Name of the categorical variable to cross-map. Default is
            "annot".

    Returns:
        If the annotations have already been cross-mapped, a pandas DataFrame
        containing the cross-tabulated annotations. Otherwise, a tuple
        containing the cross-tabulated annotations, the preprocessed AnnData
        objects used for training the logistic regression models, and the
        trained models themselves.
    """
    datasets = adata.obs[dataset].cat.categories
    if datasets.size != 2:
        raise ValueError("Number of pooled datasets != 2")
    ds1, ds2 = datasets.to_list()
    skip_prediction = (
        f"{ds1}_annot" in adata.obs.columns
        and f"{ds2}_annot" in adata.obs.columns
    )
    if not skip_prediction:
        ad1 = adata[adata.obs[dataset] == ds1].copy()
        ad2 = adata[adata.obs[dataset] == ds2].copy()
        ad1.obs[annotation].cat.rename_categories(
            lambda x: x.replace(f"{ds1}_", ""), inplace=True
        )
        ad2.obs[annotation].cat.rename_categories(
            lambda x: x.replace(f"{ds2}_", ""), inplace=True
        )
        lr1 = LR_train(ad1, annotation)
        lr2 = LR_train(ad2, annotation)
        LR_predict(adata, lr2, key_added=f"{ds2}_annot")
        LR_predict(adata, lr1, key_added=f"{ds1}_annot")
    df = pd.DataFrame(
        np.dot(
            cross_table(
                adata, f"{ds1}_annot", annotation, normalise="yx"
            ).values,
            cross_table(
                adata, annotation, f"{ds2}_annot", normalise="xy"
            ).values,
        ),
        index=adata.obs[f"{ds1}_annot"].cat.categories,
        columns=adata.obs[f"{ds2}_annot"].cat.categories,
    )
    if skip_prediction:
        return df
    else:
        return df, {ds1: ad1, ds2: ad2}, {ds1: lr1, ds2: lr2}


def auto_filter_cells(
    ad,
    min_count=250,
    min_gene=50,
    subset=True,
    filter_kw={
        "metrics": ["n_counts", "n_genes", "percent_mito"],
        "min_count": 1000,
        "min_gene": 200,
        "max_mito": 20,
        "min_pass_rate": 0.5,
        "onesided": True,
    },
) -> anndata.AnnData:
    """
    Automatically filter cells in an AnnData object based on quality control
    metrics.

    This function automatically filters cells in an AnnData object based on
    quality control metrics using a simple iterative algorithm. The algorithm
    first applies a set of default filters to the data using the
    `simple_default_pipeline` function, and then iteratively adjusts the filter
    parameters until a minimum passing rate is achieved for each metric. The
    resulting filtered AnnData object is returned with a new categorical
    variable indicating which cells passed the filters.

    Args:
        ad (anndata.AnnData): AnnData object to filter cells in.
        min_count (int): Minimum number of counts per cell to retain. Default is 250.
        min_gene (int): Minimum number of genes per cell to retain. Default is 50.
        subset (bool): Whether to return a subset of the original AnnData object
            containing only the passing cells. Default is True.
        filter_kw (dict): Dictionary of keyword arguments to pass to the
            :py:func:`simple_default_pipeline` function. Default is a set of
            default filter parameters.

    Returns:
        If `subset` is True, a new AnnData object containing only the passing
        cells. Otherwise, the original AnnData object with a new categorical
        variable indicating which cells passed the filters.

    Raises:
        QcLowPassError: If dataset does not pass minimum specifications.

    """
    filter_kw = filter_kw.copy()
    while True:
        try:
            ad1 = simple_default_pipeline(
                ad,
                filter_only=True,
                filter_kw=filter_kw,
            )
        except QcLowPassError as e:
            error_metric = str(e)
            sc_warn(f"{error_metric} passing rate too low")
            if error_metric == "n_counts":
                if not filter_kw.get("onesided", False):
                    filter_kw["onesided"] = True
                elif filter_kw["min_count"] > min_count:
                    filter_kw["min_count"] //= 2
                    sc_warn("decrease `min_count` by half")
                else:
                    ad = ad[ad.obs["n_counts"] >= min_count].copy()
                    filter_kw["metrics"].remove(error_metric)
            elif error_metric == "n_genes":
                if not filter_kw.get("onesided", False):
                    filter_kw["onesided"] = True
                elif filter_kw["min_gene"] > min_gene:
                    filter_kw["min_gene"] //= 2
                    sc_warn("decrease `min_gene` by half")
                else:
                    ad = ad[ad.obs["n_genes"] >= min_gene].copy()
                    filter_kw["metrics"].remove(error_metric)
            elif error_metric == "percent_mito" and filter_kw["max_mito"] < 50:
                filter_kw["max_mito"] += 10
                sc_warn("increase `max_mito` by 10")
            else:
                filter_kw["metrics"].remove(error_metric)
                sc_warn(f"remove {error_metric} from filters")
            continue
        break
    if subset:
        ad1.uns["auto_filter_cells"] = filter_kw
        return ad1
    else:
        ad.obs["pass_auto_filter"] = ad.obs_names.isin(ad1.obs_names)
        ad.uns["auto_filter_cells"] = filter_kw
        return ad



"""
scanpy louvain
"""

import scanpy as sc
from ._obj_utils import write_cluster


def louvain(adata, resolution, use_graph=None, key_added=None, export_cluster=None, **kwargs):
    """
    Wrapper function for sc.tl.louvain, for supporting multiple resolutions.
    """
    keys = []
    if "restrict_to" in kwargs and not (
        isinstance(kwargs["restrict_to"], (list, tuple))
        and len(kwargs["restrict_to"]) == 2
        and kwargs["restrict_to"][0]
    ):
        del kwargs["restrict_to"]
    adj_mat = None
    if use_graph:
        if use_graph not in adata.uns:
            raise KeyError(f'"{use_graph}" is not a valid key of `.uns`.')
        adj_mat = adata.uns[use_graph]["connectivities"]
    if not isinstance(resolution, (list, tuple)):
        if key_added is not None and not key_added.startswith("louvain_"):
            key_added = f"louvain_{key_added}"
        elif key_added is None:
            key_added = "louvain"
        sc.tl.louvain(
            adata, resolution=resolution, adjacency=adj_mat, key_added=key_added, **kwargs
        )
        keys.append(key_added)
    else:
        for i, res in enumerate(resolution):
            res_key = str(res).replace(".", "_")
            if key_added is None:
                graph_key = ("_" + use_graph) if use_graph else ""
                key = f"louvain{graph_key}_r{res_key}"
            elif not isinstance(key_added, (list, tuple)):
                key = f"louvain_{key_added}_r{res_key}"
            elif len(key_added) == len(resolution):
                key = key_added[i]
            else:
                raise ValueError(
                    "`key_added` can only be None, a scalar, or an "
                    "iterable of the same length as `resolution`."
                )
            keys.extend(
                louvain(
                    adata,
                    resolution=res,
                    use_graph=use_graph,
                    key_added=key,
                    **kwargs,
                )
            )

    if export_cluster:
        write_cluster(adata, keys, export_cluster)

    return keys



"""
Utility functions
"""

import warnings
import os
import re
import numpy as np
import numpy_groupies as npg
import scipy.sparse as sp
import pandas as pd
import anndata
import scanpy as sc
from ._obj_utils import (
    _set_obsm_key,
    _restore_obsm_key,
    _backup_default_key,
    _rename_default_key,
    _delete_backup_key,
)


sc_warn = sc.logging.warn if sc.__version__.startswith("1.4") else sc.logging.warning


def _find_rep(adata, use_rep, return_var_names=True):
    if isinstance(use_rep, str):
        if use_rep == "raw" and adata.raw:
            X = adata.raw.X
            var_names = adata.raw.var_names.values
        elif use_rep == "X":
            X = adata.X
            var_names = adata.var_names.values
        elif use_rep in adata.layers.keys():
            X = adata.layers[use_rep]
            var_names = adata.var_names.values
        elif use_rep in adata.obsm.keys():
            X = adata.obsm[use_rep]
            var_names = np.array([f"{use_rep}{i+1}".replace("X_", "") for i in range(X.shape[1])])
        elif use_rep in adata.obs.keys():
            X = adata.obs[use_rep].values.reshape((adata.n_obs, 1))
            var_names = np.array([use_rep])
        else:
            raise ValueError("Invalid `use_rep` provided.")
    elif isinstance(use_rep, np.ndarray) and use_rep.shape[0] == adata.n_obs:
        x = use_rep
        var_names = np.arange(x.shape[1])
    elif isinstance(use_rep, pd.DataFrame) and use_rep.shape[0] == adata.n_obs:
        x = use_rep.values
        var_names = use_rep.columns.values
    else:
        raise ValueError("Invalid `use_rep` provided.")

    if return_var_names:
        return (X, var_names)
    return X


def read_list(fn, **kwargs):
    return pd.read_csv(fn, header=None, names=["x"], **kwargs).x.to_list()


def lognorm_to_counts(X, norm_sum=1e4, n_counts=None, force=False, rounding=True):
    if not sp.issparse(X):
        X = sp.csr_matrix(X)
    X_expm1 = np.expm1(X)
    if n_counts is not None:
        size_factor = n_counts / norm_sum
        X_counts = (X_expm1.T * sp.csr_matrix(sp.diags(size_factor))).T
        res = np.abs(X_counts.data - np.round(X_counts.data)).sum() / X_counts.data.sum()
        if res < 1e-6 or force:
            if rounding:
                X_counts.data = np.round(X_counts.data).astype(np.int32)
            return X_counts
        warnings.warn(
            f"Non-integer residuals too large (res = {res}), try inferring size_factor"
        )
    x_min = np.array([X_expm1.getrow(i).data.min() for i in range(X_expm1.shape[0])])
    size_factor = 1 / x_min
    X_counts = (X_expm1.T * sp.csr_matrix(sp.diags(size_factor))).T
    res = np.abs(X_counts.data - np.round(X_counts.data)).sum() / X_counts.data.sum()
    if res < 1e-3 or force:
        if rounding:
            X_counts.data = np.round(X_counts.data).astype(np.int32)
        return X_counts
    raise ValueError(f"Non-integer residuals too large (res = {res}), failed to recover counts")


def restore_adata(
    adata,
    restore_type="count",
    use_raw=True,
    obs_cols=None,
    var_cols=None,
    obsm_keys=None,
    use_n_counts=False,
    **kwargs,
):
    if use_raw and not adata.raw:
        raise ValueError("adata.raw not found")

    if restore_type == "norm":
        if use_raw:
            X = adata.raw.X
        else:
            X = adata.X
    elif restore_type == "count":
        n_counts = None
        if use_n_counts is True:
            if use_n_counts in adata.obs_keys():
                n_counts = adata.obs[use_n_counts].values
            elif "n_counts" in adata.obs_keys():
                n_counts = adata.obs["n_counts"].values
        if use_raw:
            X0 = adata.raw.X
        else:
            X0 = adata.X
        X = lognorm_to_counts(X0, n_counts=n_counts, **kwargs)
    else:
        raise ValueError(
            f'{restore_type}: invalid <restore_type>, choose between "norm" and "count"'
        )

    obs = adata.obs[obs_cols].copy() if obs_cols else adata.obs.copy()
    if use_raw:
        var = adata.raw.var[var_cols].copy() if var_cols else adata.raw.var.copy()
    else:
        var = adata.var[var_cols].copy() if var_cols else adata.var.copy()
    ad = anndata.AnnData(X=X, obs=obs, var=var)
    if obsm_keys:
        for ok in obsm_keys:
            if ok in adata.obsm.keys():
                ad.obsm[ok] = adata.obsm[ok]
    return ad


def _get_index_in_superset(superset, x):
    index = np.argsort(superset)
    xpos = np.searchsorted(superset[index], x)
    return index[xpos]


def expand_feature_space(ad, features):
    var_names = ad.var_names
    n_features = features.size
    if not var_names.isin(features).all():
        raise ValueError("`ad.var_names` must be a strict subset of `features`")
    k_included = _get_index_in_superset(features, var_names.values)
    k_extra = np.array([i for i in range(n_features) if i not in k_included])
    new_ad = anndata.AnnData(
        X=sp.csr_matrix((ad.X.data, ad.X.indices, ad.X.indptr), shape=(ad.n_obs, n_features)),
        obs=ad.obs,
    )
    new_ad.var_names = var_names.tolist() + list(features[k_extra])
    for layer in ad.layers.keys():
        layer_X = ad.layers[layer]
        new_ad.layers[layer] = sp.csr_matrix(
            (layer_X.data, layer_X.indices, layer_X.indptr), shape=(ad.n_obs, n_features)
        )
    return new_ad[:, features].copy()


def find_top_expressed_genes(ad, use_rep="X", groupby=None, n_genes=50, inplace=True):
    X, var_names = _find_rep(ad, use_rep)
    counts = X.sum(axis=1)
    if sp.issparse(X):
        counts = counts.A1
    X = sc.preprocessing._normalization._normalize_data(X, counts, after=100, copy=True)
    if groupby:
        raise NotImplementedError
    else:
        mean_percent = X.mean(axis=0)
        if sp.issparse(X):
            mean_percent = mean_percent.A1
        top_idx = np.argsort(mean_percent)[::-1][:n_genes]
    top_genes = var_names[top_idx]
    if inplace:
        ad.var[f"top{n_genes}"] = ad.var_names.isin(top_genes)
    else:
        return top_genes


def remove_genes(adata, var_flags):
    if isinstance(var_flags, (tuple, list)):
        pass
    elif isinstance(var_flags, str):
        var_flags = [var_flags]
    else:
        var_flags = []
    if var_flags:
        mask = np.zeros(adata.n_vars).astype(bool)
        for vf in var_flags:
            mask = mask | adata.var[vf]
        adata = adata[:, ~mask].copy()
    return adata


def project_into_PC(
    source_adata,
    target_adata,
    source_scaled=False,
    target_scaled=True,
    source_rep="X",
    target_pca="X_pca",
    use_loadings=None,
    key_added=None,
):
    """project cells from one dataset (source) into PC space of another dataset (target)"""
    if source_rep == "raw":
        src_ad = source_adata.raw.to_anndata()
    else:
        src_ad = source_adata

    tgt_ad = target_adata

    if np.any(~tgt_ad.var_names.isin(src_ad.var_names)):
        raise ValueError("`source_adata` must contain all features in `target_adata`")

    if np.any(src_ad.var_names[src_ad.var_names.isin(tgt_ad.var_names)] != tgt_ad.var_names):
        raise ValueError("features in `source_adata` must be in the same order as `target_adata`")

    tgt_X = tgt_ad.X
    tgt_X_is_dense = not sp.issparse(tgt_X)

    aux_ad = src_ad[:, src_ad.var_names.isin(tgt_ad.var_names)].copy()

    if not source_scaled:
        sc.pp.scale(aux_ad, zero_center=tgt_X_is_dense, max_value=10)

    if not target_scaled:
        tgt_X = sc.pp.scale(tgt_X, zero_center=tgt_X_is_dense, max_value=10)

    if use_loadings:
        tgt_V = use_loadings
    else:
        if not tgt_X_is_dense:
            tgt_X = tgt_X.toarray()
        tgt_V = np.linalg.lstsq(tgt_X, tgt_ad.obsm[target_pca])

    proj_pca = aux_ad.X @ tgt_V[0]
    del aux_ad, src_ad

    if key_added:
        source_adata.obsm[f"X_pca_{key_added}"] = proj_pca
        return tgt_V
    return proj_pca, tgt_V


def _is_numeric(x):
    return x.dtype.kind in ("i", "f")


def cross_table(
    adata,
    x,
    y,
    normalise=None,
    exclude_x=None,
    exclude_y=None,
    subset=None,
    sort=False,
    include_nan=False,
):
    """Make a cross table comparing two categorical annotations"""
    x_attr = adata.obs[x]
    y_attr = adata.obs[y]
    assert not _is_numeric(x_attr.values), f"Can not operate on numerical {x}"
    assert not _is_numeric(y_attr.values), f"Can not operate on numerical {y}"
    if include_nan:
        x_attr = x_attr.astype(str).astype("category")
        y_attr = y_attr.astype(str).astype("category")
    if subset is not None:
        x_attr = x_attr[subset]
        y_attr = y_attr[subset]
    selection = np.ones(x_attr.size, dtype=bool)
    if exclude_x:
        selection = selection & ~x_attr.isin(exclude_x)
    if exclude_y:
        selection = selection & ~y_attr.isin(exclude_y)
    x_attr = x_attr[selection]
    y_attr = y_attr[selection]

    if normalise in (0, "x", "xy", "row", "index"):
        norm_method = "index"
    elif normalise in (1, "y", "yx", "col", "columns"):
        norm_method = "columns"
    else:
        norm_method = False
    crs_tbl = pd.crosstab(x_attr, y_attr, dropna=False, normalize=norm_method)
    nx, ny = crs_tbl.shape
    if normalise in ("xy", "yx", "jaccard"):
        x_sizes = crs_tbl.sum(axis=1).values
        y_sizes = crs_tbl.sum(axis=0).values
        if normalise == "yx":
            normaliser = np.tile(x_sizes.reshape(nx, 1), (1, ny))
        elif normalise == "xy":
            normaliser = np.tile(y_sizes.reshape(1, ny), (nx, 1))
        else:
            normaliser = (
                np.tile(x_sizes.reshape(nx, 1), (1, ny))
                + np.tile(y_sizes.reshape(1, ny), (nx, 1))
                - crs_tbl.values
            )
        crs_tbl = (crs_tbl / normaliser).round(4)
    if sort in ("x", "index", "xy", "yx", "both"):
        crs_tbl = crs_tbl.sort_index()
    if sort in ("y", "columns", "xy", "yx", "both"):
        crs_tbl = crs_tbl.sort_columns()
    return crs_tbl


def run_celltypist(
    ad, model, use_rep="X", require_lognorm=False, min_prob=None, key_added="ctp_pred"
):
    import celltypist

    X, var_names = _find_rep(ad, use_rep)
    aux_ad = anndata.AnnData(X=X)
    aux_ad.obs_names = ad.obs_names
    aux_ad.var_names = var_names
    if require_lognorm:
        sc.pp.normalize_total(aux_ad, target_sum=1e4)
        sc.pp.log1p(aux_ad)
    pred = celltypist.annotate(aux_ad, model=model, majority_voting=True)
    ad.obs[f"{key_added}"] = pred.predicted_labels.majority_voting
    ad.obs[f"{key_added}_prob"] = pred.probability_matrix.max(axis=1)
    if min_prob is not None:
        ad.obs[f"{key_added}_uncertain"] = ad.obs[f"{key_added}_prob"] < min_prob
        ad.obs[f"{key_added}_uncertain"] = ad.obs[f"{key_added}_uncertain"].astype("category")
    del aux_ad


def run_harmony(adata, batch, theta=2.0, use_rep="X_pca", key_added="hm", random_state=0, **kwargs):
    if not isinstance(batch, (tuple, list)):
        batch = [batch]
    if not isinstance(theta, (tuple, list)):
        theta = [theta] * len(batch)
    for b in batch:
        if b not in adata.obs.columns:
            raise KeyError(f"{b} is not a valid obs annotation.")
    if use_rep not in adata.obsm.keys():
        raise KeyError(f"{use_rep} is not a valid embedding.")
    meta = adata.obs[batch].reset_index()
    embed = adata.obsm[use_rep]

    if sc.__version__.startswith("1.4"):
        # ===========
        import rpy2.robjects
        from rpy2.robjects.packages import importr

        harmony = importr("harmony")
        from rpy2.robjects import numpy2ri, pandas2ri

        numpy2ri.activate()
        pandas2ri.activate()
        set_seed = rpy2.robjects.r("set.seed")
        set_seed(random_state)
        hm_embed = harmony.HarmonyMatrix(
            embed, meta, batch, theta, do_pca=False, verbose=False, **kwargs
        )
        pandas2ri.deactivate()
        numpy2ri.deactivate()
        hm_embed = numpy2ri.ri2py(hm_embed)
        # ===========
    else:
        import harmonypy as hm

        if "max_iter_harmony" not in kwargs:
            kwargs["max_iter_harmony"] = 20
        hobj = hm.run_harmony(embed, meta, batch, theta=theta, **kwargs)
        hm_embed = hobj.Z_corr.T

    if key_added:
        adata.obsm[f"{use_rep}_{key_added}"] = hm_embed
    else:
        adata.obsm[use_rep] = hm_embed.T


def run_bbknn(adata, batch, use_rep="X_pca", key_added="bk", **kwargs):
    import bbknn

    _set_obsm_key(adata, "X_pca", use_rep)
    try:
        _backup_default_key(adata.uns, "neighbors")
        bbknn.bbknn(adata, batch_key=batch, **kwargs)
        if key_added:
            _rename_default_key(adata.uns, "neighbors", f"neighbors_{key_added}")
            if not sc.__version__.startswith("1.4"):
                _rename_default_key(adata.obsp, "distances", f"neighbors_{key_added}_distances")
                _rename_default_key(
                    adata.obsp, "connectivities", f"neighbors_{key_added}_connectivities"
                )
                adata.uns[f"neighbors_{key_added}"][
                    "distances_key"
                ] = f"neighbors_{key_added}_distances"
                adata.uns[f"neighbors_{key_added}"][
                    "connectivities_key"
                ] = f"neighbors_{key_added}_connectivities"
        else:
            _delete_backup_key(adata.uns, "neighbors")
    finally:
        _restore_obsm_key(adata, "X_pca", use_rep)


def run_phate(
    adata,
    use_rep="X",
    key_added=None,
    knn=5,
    decay=40,
    t="auto",
    n_pca=100,
    random_state=0,
    verbose=False,
    **kwargs,
):
    import phate

    try:
        data = _find_rep(adata, use_rep, return_var_names=False)
    except ValueError as e:
        if use_rep in adata.uns.keys():
            data = adata.uns[use_rep]["distances"]
            kwargs["knn_dist"] = "precomputed"
        else:
            raise e
    phate_operator = phate.PHATE(
        knn=knn, decay=decay, t=t, n_pca=n_pca, random_state=random_state, verbose=verbose, **kwargs
    )
    kc_phate = phate_operator.fit_transform(data)
    slot_name = f"X_phate_{key_added}" if key_added else "X_phate"
    adata.obsm[slot_name] = kc_phate


def run_diffQuant(
    adata,
    groupby,
    condition,
    ctrl,
    outprfx,
    trmt=None,
    groups=None,
    cofactor=[],
    n_cpu=1,
    recover_counts=True,
    dry=False,
    debug=False,
):
    if not isinstance(cofactor, (tuple, list)):
        cofactor = [cofactor]
    selected_variables = [groupby, condition]
    selected_variables.extend(cofactor)

    X = adata.raw.X.copy() if adata.raw else adata.X

    if recover_counts:
        X = lognorm_to_counts(
            X, adata.obs["n_counts"].values, rounding=True, force=(recover_counts == "force")
        )
    else:
        X.data = np.round(np.expm1(X.data))

    if groups is not None and isinstance(groups, (list, tuple)):
        k_obs = adata.obs[groupby].isin(groups)
        ad = anndata.AnnData(X=X[k_obs, :], obs=adata.obs.loc[k_obs, selected_variables].copy())
    else:
        ad = anndata.AnnData(X=X, obs=adata.obs[selected_variables].copy())
    ad.var_names = adata.raw.var_names if adata.raw else adata.var_names
    ad.obs[groupby].cat.remove_unused_categories(inplace=True)

    count_matrix = pseudo_bulk(ad, groupby=groupby, FUN=np.sum).astype(np.int32)
    metadata = ad.obs.groupby(groupby).first()
    metadata.index.name = "sample"

    import tempfile

    x_fh = tempfile.NamedTemporaryFile(suffix=".tsv", delete=False)
    meta_fh = tempfile.NamedTemporaryFile(suffix=".tsv", delete=False)

    count_matrix.to_csv(x_fh.name, sep="\t")
    metadata.to_csv(meta_fh.name, sep="\t")

    import subprocess as sbp

    trmt_opt = f"--trmt {trmt}" if trmt else ""
    covar_opt = f'--covar {",".join(cofactor)}' if cofactor else ""
    cmd = (
        f"diffQuant -o {outprfx} --minfrac 0.99 --nCPU {n_cpu} --factor {condition} "
        f"--ctrl {ctrl} {trmt_opt} {covar_opt} {x_fh.name} {meta_fh.name}"
    )
    if dry:
        print(cmd)
        return
    try:
        job = sbp.run(cmd, shell=True, capture_output=True)
    except sbp.CalledProcessError as e:
        print(e)
    finally:
        if debug:
            print(job.stdout.decode("utf-8"))
            print(job.stderr.decode("utf-8"))

        os.remove(x_fh.name)
        os.remove(meta_fh.name)
    de_tbl = pd.read_csv(f"{outprfx}.{trmt}-vs-{ctrl}.txt", sep="\t", index_col=0)
    return de_tbl


def run_cellphonedb(
    adata,
    groupby,
    group_size=-1,
    cpdb_path="cellphonedb",
    outpath="./cpdb",
    thread=8,
    dry=False,
    **subsample_kwargs,
):
    outdir, outname = os.path.split(outpath)
    if not outdir:
        outdir = "."
    os.makedirs(outdir, exist_ok=True)

    import tempfile

    tmp_h5ad = tempfile.NamedTemporaryFile(suffix=".h5ad", delete=False)
    if group_size > 0:
        adata = subsample(
            adata,
            fraction=1,
            groupby=groupby,
            max_n=group_size,
            min_n=group_size,
            **subsample_kwargs,
        )
    adata.write(tmp_h5ad.name, compression="lzf")

    tmp_obs = f"{outpath}.obs.tsv"
    adata.obs[[groupby]].to_csv(tmp_obs, sep="\t")

    import subprocess as sbp

    cmd = (
        f"{cpdb_path} method statistical_analysis --counts-data gene_name --output-format tsv"
        f" --project-name {outname} --output-path {outdir} --threads {thread}"
        f" {tmp_obs} {tmp_h5ad.name}"
    )
    try:
        if dry:
            print(cmd)
            return
        else:
            sbp.run(cmd.split())
    except sbp.CalledProcessError as e:
        print(e)
    finally:
        os.remove(tmp_h5ad.name)


def run_cellchat(
    adata,
    groupby,
    use_rep="raw",
    organism="human",
):

    X, var_names = _find_rep(adata, use_rep)

    import rpy2.robjects as ro
    from rpy2.robjects.packages import importr
    from rpy2.robjects import numpy2ri, pandas2ri
    from rpy2.robjects.conversion import localconverter

    rBase = importr("base")
    rCC = importr("CellChat")
    rMatrix = importr("Matrix")
    rPbapply = importr("pbapply")
    rPbapply.pboptions(type="none")

    cellchat_db = ro.r(f"CellChatDB.{organism}")

    if sp.issparse(X):
        X = X.tocoo()
        r_X = rMatrix.sparseMatrix(
            i=numpy2ri.py2rpy(X.col + 1),
            j=numpy2ri.py2rpy(X.row + 1),
            x=rBase.as_numeric(numpy2ri.py2rpy(X.data)),
            dims=rBase.as_integer(list(X.shape[::-1])),
            dimnames=rBase.list(
                rBase.as_character(numpy2ri.py2rpy(var_names)),
                rBase.as_character(numpy2ri.py2rpy(adata.obs_names.values)),
            ),
        )
    else:
        r_X = rBase.matrix(
            numpy2ri.py2rpy(X),
            dimnames=rBase.list(
                rBase.as_character(numpy2ri.py2rpy(var_names)),
                rBase.as_character(numpy2ri.py2rpy(adata.obs_names.values)),
            ),
        )

    with localconverter(ro.default_converter + pandas2ri.converter):
        r_obs = ro.conversion.py2rpy(adata.obs[[groupby]])

    cellchat = rCC.createCellChat(object=r_X, meta=r_obs, group_by=groupby)

    cellchat_db_secret = rCC.subsetDB(cellchat_db, search="Secreted Signaling")
    cellchat.slots["DB"] = cellchat_db_secret
    cellchat = rCC.subsetData(cellchat)
    cellchat = rCC.identifyOverExpressedGenes(cellchat)
    cellchat = rCC.identifyOverExpressedInteractions(cellchat)
    cellchat = rCC.computeCommunProb(cellchat)
    cellchat = rCC.filterCommunication(cellchat, min_cells=10)
    cellnet = rCC.subsetCommunication(cellchat)

    cellnet = pandas2ri.rpy2py(cellnet)
    return cellnet


def run_celltype_composition_analysis(
    adata,
    sample,
    celltype,
    cat_covar=[],
    num_covar=[],
    extra_term=None,
    outprfx=None,
):
    import rpy2.robjects as ro
    from rpy2.robjects.packages import importr
    from rpy2.robjects import numpy2ri, pandas2ri
    from rpy2.robjects.conversion import localconverter

    rBase = importr("base")
    rUtils = importr("utils")
    rMatrix = importr("Matrix")
    rTibble = importr("tibble")
    rSCTKR = importr("sctkr")

    with localconverter(ro.default_converter + pandas2ri.converter):
        r_obs = ro.conversion.py2rpy(adata.obs)

    ctca_result = rSCTKR.CellTypeCompositionAnalysis(
        rTibble.as_tibble(r_obs),
        sample,
        celltype,
        cat_covar,
        num_covar,
        extra_term=extra_term,
        save=outprfx,
    )

    return ctca_result


def _run_NMF(args, X=None, kwargs=None):
    from sklearn.decomposition import non_negative_factorization

    if not isinstance(kwargs, dict):
        kwargs = {}
    k, seed = args
    _, H, _ = non_negative_factorization(X, n_components=k, random_state=seed, **kwargs)
    return H


def run_cNMF(
    adata,
    K,
    use_rep="X",
    rep_type="count",
    n_iter=100,
    random_state=0,
    density_threshold=2,
    n_worker=1,
    use_cache=False,
    output_prefix="cNMF",
    **kwargs,
):
    """cNMF as in Kotliar et al, re-implemented by nh3"""
    if isinstance(K, int):
        K = [K]
    K = np.array(K).astype(int)

    default_NMF_kwargs = {
        "init": "random",
        "solver": "cd",
        "beta_loss": "frobenius",
        "tol": 1e-4,
        "max_iter": 1000,
        "alpha_W": 0.0,
        "alpha_H": "same",
        "l1_ratio": 0.0,
    }
    for k, v in default_NMF_kwargs.items():
        kwargs[k] = kwargs.get(k, v)

    if "highly_variable" not in adata.var:
        raise ValueError("`adata` must have highly variable genes identified.")
    hvg = adata.var_names[adata.var.highly_variable.values]

    X, var_names = _find_rep(adata, use_rep)
    var_names = pd.Index(var_names)

    if not sp.issparse(X):
        raise ValueError("Provided X must be sparse")

    if rep_type not in ("count", "lognorm"):
        raise ValueError("`rep_type` must be 'count' or 'lognorm'.")

    # import tempfile
    from functools import partial
    from multiprocessing.pool import Pool
    import sklearn
    from sklearn.decomposition import non_negative_factorization
    from sklearn.metrics import pairwise_distances, silhouette_score
    from sklearn.cluster import KMeans

    if rep_type == "count":
        tpm_X = sklearn.preprocessing.normalize(X, norm="l1", axis=1) * 1e6
    else:
        tpm_X = np.expm1(X) * 1e2
    norm_X = sklearn.preprocessing.scale(tpm_X, with_mean=False, axis=0)
    if np.all(var_names.isin(hvg)):
        norm_X_hvg = norm_X
    else:
        norm_X_hvg = norm_X[:, var_names.isin(hvg)]

    n_runs = K.size * n_iter
    np.random.seed(seed=random_state)
    nmf_seeds = np.random.randint(low=1, high=(2 ** 32) - 1, size=n_runs)
    n_neighbors = int(0.3 * n_iter)

    run_NMF = partial(_run_NMF, X=norm_X_hvg, kwargs=kwargs)

    for i, k in enumerate(K):
        spectral_cache = f"{output_prefix}.spectral.k{k}.combined.csv.gz"
        consensus_spectra_csv = f"{output_prefix}.spectra.k_{k}.consensus.csv.gz"
        consensus_usage_csv = f"{output_prefix}.usage.k_{k}.consensus.csv.gz"
        consensus_stats_csv = f"{output_prefix}.stats.k_{k}.consensus.csv"
        gene_spectra_csv = f"{output_prefix}.gene_spectra.k_{k}.csv.gz"
        gene_spectra_tpm_csv = f"{output_prefix}.gene_spectra_tpm.k_{k}.csv.gz"

        if use_cache:
            print(f"use cache {spectral_cache}")
            combined_spectra = pd.read_csv(spectral_cache, index_col=0)
        else:
            with Pool(n_worker) as pool:
                tasks = ((k, seed) for seed in nmf_seeds[i * n_iter : (i + 1) * n_iter])
                results = pool.map(run_NMF, tasks)
                # `combined_spectra` is a (k * n_inter, n_gene) dataframe
                combined_spectra = pd.DataFrame(
                    np.concatenate(results, axis=0),
                    index=[f"iter{n}_topic{m}" for n in range(n_iter) for m in range(k)],
                    columns=var_names[var_names.isin(hvg)],
                )
                combined_spectra.to_csv(spectral_cache)

        l2_spectra = (combined_spectra.T / np.linalg.norm(combined_spectra.values, 2, axis=1)).T
        topics_dist = pairwise_distances(l2_spectra.values)
        partition_order = np.argpartition(topics_dist, n_neighbors + 1)[:, : n_neighbors + 1]
        distance_to_nearest_neighbors = topics_dist[
            np.arange(topics_dist.shape[0])[:, None], partition_order
        ]
        local_density = pd.Series(
            distance_to_nearest_neighbors.sum(1) / (n_neighbors),
            index=l2_spectra.index,
        )
        del partition_order, distance_to_nearest_neighbors
        density_filter = local_density.values < density_threshold
        l2_spectra = l2_spectra.loc[density_filter, :]

        kmeans_model = KMeans(n_clusters=k, n_init=10, random_state=1)
        kmeans_cluster_labels = pd.Series(
            kmeans_model.fit_predict(l2_spectra) + 1, index=l2_spectra.index
        )
        median_spectra = l2_spectra.groupby(kmeans_cluster_labels).median()
        median_spectra = (median_spectra.T / median_spectra.sum(axis=1)).T
        median_spectra.to_csv(consensus_spectra_csv)

        stability = silhouette_score(l2_spectra.values, kmeans_cluster_labels, metric="euclidean")

        refit_kwargs = kwargs.copy()
        refit_kwargs.update(dict(H=median_spectra.values.astype(np.float32), update_H=False))
        refit_W, _, _ = non_negative_factorization(
            norm_X_hvg, n_components=k, random_state=1, **refit_kwargs
        )
        rf_usage = pd.DataFrame(
            refit_W,
            index=adata.obs_names,
            columns=median_spectra.index,
        )
        rf_usage.to_csv(consensus_usage_csv)

        rf_pred_norm_counts = rf_usage.dot(median_spectra)
        prediction_error = ((norm_X_hvg.toarray() - rf_pred_norm_counts) ** 2).values.sum()

        consensus_stats = pd.DataFrame(
            [k, density_threshold, stability, prediction_error],
            index=["k", "local_density_threshold", "stability", "prediction_error"],
            columns=["stats"],
        )
        consensus_stats.to_csv(consensus_stats_csv)

        gene_spectra = pd.DataFrame(
            sp.csr_matrix(np.linalg.pinv(rf_usage.values)).dot(norm_X).toarray(),
            index=rf_usage.columns,
            columns=var_names,
        )
        gene_spectra.to_csv(gene_spectra_csv)

        norm_usage = sklearn.preprocessing.normalize(rf_usage, norm="l1", axis=0)
        refit_kwargs.update(dict(H=norm_usage.T))
        refit_tpm_W, _, _ = non_negative_factorization(
            tpm_X.T, n_components=k, random_state=1, **refit_kwargs
        )
        gene_spectra_tpm = pd.DataFrame(
            refit_tpm_W,
            index=adata.var_names,
            columns=median_spectra.index,
        )
        gene_spectra_tpm.to_csv(gene_spectra_tpm_csv)
        print(f"k = {k} done")


def write_10x_h5(
    adata,
    outfile,
    data_type="gex",
    genome="GRCh38",
    feature_id=None,
    feature_name=None,
    feature_type=None,
    target_sets=None,
    other_feature_tags=None,
    data_version=2,
    software_version="3.0.0",
):
    """Write an AnnData to an HDF5 that Cellranger produces
    * Parameters
        + adata : AnnData
        An AnnData object
        + outfile : str
        Name of the output h5
        + data_type : str
        Data type, can be "gex" or "atac", default "gex"
        + genome : str
        Genome name, default "GRCh38"
        + feature_id : str
        Column in `var` that contains feature ids, by default "gene_ids" for
        "gex" or "peak" for "atac"
        + feature_name : str
        Column in `var` that contains feature names, by default `var_names`
        + feature_type : str
        Either column in `var` that contains feature types, or the actual
        feature type common to all features, by default "Gene Expression" for
        "gex" or "Peaks" for "atac"
        + target_sets : list of str
        A list of boolean columns in `var` that flag targeted features, only
        applicable to targeted gene expression assay
        + other_feature_tags : list of str
        Other columns in `var` to save
        + data_version : numeric
        Version of the assay, default 2
        + software_version : str
        Cellranger version, default "3.0.0"
    """
    import h5py

    DEFAULT_FEATURE_IDS = {"gex": "gene_ids", "atac": "peak"}
    DEFAULT_FEATURE_TYPES = {"gex": "Gene Expression", "atac": "Peaks"}
    DEFAULT_TAG_KEYS = {"gex": ["genome"], "atac": ["genome", "derivation"]}

    other_feature_tags = other_feature_tags or []

    n_barcode, n_feature = adata.shape

    barcode_len = max(map(len, adata.obs_names.tolist()))

    feature_type = feature_type or DEFAULT_FEATURE_TYPES[data_type]
    ftype = (
        adata.var.get(feature_type).values
        if feature_type in adata.var.columns
        else np.repeat(feature_type, n_feature)
    )
    ftype_len = max(map(len, ftype.tolist()))

    fname = adata.var.get(feature_name, adata.var_names).values
    fname_len = max(map(len, fname.tolist()))

    fid = adata.var.get(feature_id or DEFAULT_FEATURE_IDS[data_type], adata.var_names).values
    fid_len = max(map(len, fid.tolist()))

    nodes = {
        "/matrix": (h5py.Group, None),
        "/matrix/barcodes": (f"S{barcode_len}", adata.obs_names.values),
        "/matrix/data": (np.int32, adata.X.data),
        "/matrix/indices": (np.int32, adata.X.indices),
        "/matrix/indptr": (np.int32, adata.X.indptr),
        "/matrix/shape": (np.int64, np.array([n_feature, n_barcode])),
        "/matrix/features": (h5py.Group, None),
        "/matrix/features/feature_type": (f"S{ftype_len}", ftype),
        "/matrix/features/id": (f"S{fid_len}", fid),
        "/matrix/features/name": (f"S{fname_len}", fname),
        "/matrix/features/genome": (f"S{len(genome)}", np.repeat(genome, n_feature)),
    }

    if data_type == "atac":
        if "derivation" in adata.var.columns:
            other_feature_tags.append("derivation")
        else:
            nodes["/matrix/features/derivation"] = ("S1", np.repeat("", n_feature))

    if data_type == "gex" and target_sets is not None:
        if isinstance(target_sets, str):
            target_sets = [target_sets]
        valid_target_sets = []
        for tgt_set in target_sets:
            if adata.var.dtypes.get(tgt_set) == "bool":
                valid_target_sets.append(tgt_set)
            else:
                warnings.warn(f"Ignore invalid target set: {tgt_set}")
        if valid_target_sets:
            nodes["/matrix/features/target_sets"] = (h5py.Group, None)
            for tgt_set in valid_target_sets:
                nodes[f"/matrix/features/target_sets/{tgt_set}"] = (
                    np.int32,
                    np.where(adata.var.get(tgt_set).values)[0],
                )

    tag_keys_set = set(DEFAULT_TAG_KEYS.get(data_type, []))
    for ft in other_feature_tags:
        if ft in adata.var.columns:
            path = f"/matrix/features/{ft}"
            dv = adata.var.get(ft).values
            dt = dv.dtype
            if dt.kind == "O":
                dl = max(map(len, dv.tolist()))
                dt = f"S{dl}"
            nodes[path] = (dt, dv)
            tag_keys_set.add(ft)
    tag_keys = np.array(list(tag_keys_set))
    tag_len = max(map(len, tag_keys.tolist()), default=1)
    nodes["/matrix/features/_all_tag_keys"] = (f"S{tag_len}", tag_keys)

    with h5py.File(outfile, mode="w") as f:
        f.attrs["filetype"] = "matrix"
        f.attrs["software_version"] = software_version
        f.attrs["version"] = data_version
        for path, node in nodes.items():
            print(path)
            dtype, dvalue = node
            if dtype == h5py.Group:
                if path not in f:
                    f.create_group(path)
            else:
                f.create_dataset(path, data=dvalue.astype(dtype), dtype=dtype, compression="gzip")


def write_mtx(
    adata,
    fname_prefix="",
    var=["gene_ids"],
    obs=None,
    use_raw=False,
    output_version="v3",
    feature_type="Gene Expression",
):
    """Export AnnData object to mtx formt
    * Parameters
        + adata : AnnData
        An AnnData object
        + fname_prefix : str
        Prefix of the exported files. If not empty and not ending with '/' or '_',
        a '_' will be appended. Full names will be <fname_prefix>matrix.mtx(.gz),
        <fname_prefix>genes.tsv/(features.tsv.gz), <fname_prefix>barcodes.tsv(.gz)
        + var : list
        A list of extra column names to be exported to gene table, default ["gene_ids"]
        + obs : list
        A list of extra column names to be exported to barcode/cell table, default None
        + use_raw: boolean
        Whether to write `adata.raw` instead of `adata`, default False
        + output_version: str
        Write v2 or v3 Cellranger mtx outputs, default v3
        + feature_type: str
        Text added as the last column of "features.tsv.gz", only relevant when
        `output_version="v3"`
    """
    if fname_prefix and not (fname_prefix.endswith("/") or fname_prefix.endswith("_")):
        fname_prefix = fname_prefix + "_"
    if var is None:
        var = []
    if obs is None:
        obs = []
    adata_obs = adata.obs.copy()
    if use_raw:
        adata = adata.raw
    obs = list(set(obs) & set(adata_obs.columns))
    var = list(set(var) & set(adata.var.columns))

    mat = sp.coo_matrix(adata.X)
    n_obs, n_var = mat.shape
    n_entry = len(mat.data)
    header = "%%MatrixMarket matrix coordinate real general\n%\n{} {} {}\n".format(
        n_var, n_obs, n_entry
    )
    df = pd.DataFrame({"col": mat.col + 1, "row": mat.row + 1, "data": mat.data})
    if output_version == "v2":
        mtx_fname = fname_prefix + "matrix.mtx"
        gene_fname = fname_prefix + "genes.tsv"
        barcode_fname = fname_prefix + "barcodes.tsv"
    else:
        mtx_fname = fname_prefix + "matrix.mtx.gz"
        gene_fname = fname_prefix + "features.tsv.gz"
        barcode_fname = fname_prefix + "barcodes.tsv.gz"
    with open(mtx_fname, "a", encoding="utf8") as fh:
        fh.write(header)
        df.to_csv(fh, sep=" ", header=False, index=False)

    obs_df = adata_obs[obs].reset_index(level=0)
    obs_df.to_csv(barcode_fname, sep="\t", header=False, index=False)
    var_df = adata.var[var].reset_index(level=0)
    if not var:
        var_df["gene"] = var_df["index"]
    if output_version != "v2":
        var_df["feature_type"] = feature_type
    var_df.to_csv(gene_fname, sep="\t", header=False, index=False)


def write_table(
    adata,
    outdir=".",
    slots=("X", "obs", "var", "obsm", "varm", "raw.X", "raw.var"),
    fmt="tsv",
    transpose=False,
    compression=False,
    X_dtype=None,
    raw_X_dtype=None,
    obs_columns=None,
    var_columns=None,
    obsm_keys=None,
    varm_keys=None,
    raw_var_columns=None,
):
    if not outdir:
        raise ValueError("`outdir` cannot be empty")
    outdir = outdir.rstrip("/")
    if not os.path.exists(outdir):
        os.mkdir(outdir)
    sep = "," if fmt == "csv" else "\t"
    suffix = ".gz" if compression else ""

    for slot in slots:
        if slot in ("X", "raw.X"):
            ad = adata.raw if slot == "raw.X" else adata
            dtype = raw_X_dtype if slot == "raw.X" else X_dtype
            if dtype is None:
                dtype = np.float32
            X = ad.X.T if transpose else ad.X
            if sp.issparse(X):
                X = X.toarray()
            if dtype in (int, np.int16, np.int32, np.int64):
                X = np.round(X)
            if transpose:
                df = pd.DataFrame(
                    X, index=ad.var_names.values, columns=adata.obs_names.values, dtype=dtype
                )
                df.index.name = "Gene"
            else:
                df = pd.DataFrame(
                    X, index=adata.obs_names.values, columns=ad.var_names.values, dtype=dtype
                )
                df.index.name = "Cell"
            df.to_csv(f"{outdir}/{slot}.{fmt}{suffix}", sep=sep)
        elif slot in ("obs", "var") and getattr(adata, slot) is not None:
            df_columns = obs_columns if slot == "obs" else var_columns
            if df_columns is None:
                df_columns = df.columns.to_list()
            elif isinstance(df_columns, str):
                df_columns = [df_columns]
            df = getattr(adata, slot)[df_columns]
            df.index.name = "Cell" if slot == "obs" else "Gene"
            df.to_csv(f"{outdir}/{slot}.{fmt}{suffix}", sep=sep)
        elif slot == "raw.var" and adata.raw is not None and adata.raw.var is not None:
            df_columns = raw_var_columns
            if df_columns is None:
                df_columns = df.columns.to_list()
            elif isinstance(df_columns, str):
                df_columns = [df_columns]
            df = adata.raw.var[df_columns]
            df.index.name = "Gene"
            df.to_csv(f"{outdir}/{slot}.{fmt}{suffix}", sep=sep)
        elif slot in ("obsm", "varm") and getattr(adata, slot) is not None:
            obj_keys = obsm_keys if slot == "obsm" else varm_keys
            if obj_keys is None:
                obj_keys = obj.keys()
            elif isinstance(df_columns, str):
                df_columns = [df_columns]
            obj = getattr(adata, slot)
            for k in obj_keys:
                X = obj[k]
                ind = adata.obs_names.values if slot == "obsm" else adata.var_names.values
                col = [f"{k}_{i}" for i in range(X.shape[1])]
                df = pd.DataFrame(X, index=ind, columns=col, dtype=np.float32)
                df.index.name = "Cell" if slot == "obsm" else "Gene"
                df.to_csv(f"{outdir}/{slot}.{k}.{fmt}{suffix}", sep=sep)


def split_by_group(adata, groupby, out_type="dict"):
    if groupby not in adata.obs.columns:
        raise KeyError(f"{groupby} is not a valid obs annotation.")
    groups = sorted(list(adata.obs[groupby].unique()))
    if out_type == "dict":
        out_adatas = {}
        for grp in groups:
            out_adatas[grp] = adata[adata.obs[groupby] == grp, :].copy()
    elif out_type == "list":
        out_adatas = []
        for grp in groups:
            out_adatas.append(adata[adata.obs[groupby] == grp, :].copy())
    else:
        raise ValueError(f'{out_type}: unsupported type, choose from "dict" or "list".')
    return out_adatas


def show_obs_categories(ad, columns=None):
    if columns:
        columns = [k for k in columns if k in ad.obs.columns]
    else:
        columns = [k for k in ad.obs.columns if ad.obs[k].dtype.name == "category"]
    for k in columns:
        summary = ad.obs[k].value_counts(dropna=False, sort=False)
        print(f"{k}, {len(summary.to_dict())}")
        print(summary)
        print("\n")


def regroup(adata, groupby, regroups):
    if groupby not in adata.obs.columns:
        raise KeyError(f"{groupby} is not a valid obs annotation.")
    groups = adata.obs[groupby].astype(str)
    new_groups = groups.copy()
    for new_grp, old_grps in regroups.items():
        if isinstance(old_grps, (list, tuple)):
            for grp in old_grps:
                new_groups[groups == grp] = new_grp
        else:
            new_groups[groups == old_grps] = new_grp
    regroup_keys = [g for g in regroups.keys() if g in set(new_groups.unique())]
    new_groups = new_groups.astype("category")
    if len(regroup_keys) == len(new_groups.cat.categories):
        new_groups = new_groups.cat.reorder_categories(regroup_keys)
    return new_groups


def subsample(
    adata,
    fraction,
    groupby=None,
    min_n=0,
    max_n=10000,
    method="random",
    index_only=False,
    random_state=0,
):
    if method not in ("random", "top"):
        raise NotImplementedError(f"method={method} unsupported")
    if groupby:
        if groupby not in adata.obs.columns:
            raise KeyError(f"{groupby} is not a valid obs annotation.")
        groups = adata.obs[groupby].unique()
        sampled_obs_names = []
        for grp in groups:
            k = adata.obs[groupby] == grp
            grp_size = sum(k)
            ds_grp_size = int(min(max_n, max(np.ceil(grp_size * fraction), min(min_n, grp_size))))
            if method == "top":
                idx = np.argsort(-adata.obs.loc[k, "n_counts"].values)[0:ds_grp_size]
            else:
                np.random.seed(random_state)
                idx = np.random.choice(grp_size, ds_grp_size, replace=False)
            sampled_obs_names.extend(list(adata.obs_names[k][idx]))
    else:
        ds_size = int(adata.n_obs * fraction)
        np.random.seed(random_state)
        idx = np.random.choice(adata.n_obs, ds_size, replace=False)
        sampled_obs_names = adata.obs_names[idx]

    if index_only:
        return sampled_obs_names
    return adata[adata.obs_names.isin(sampled_obs_names)].copy()


def random_partition(
    adata,
    partition_size,
    groupby=None,
    method="random_even",
    key_added="partition_labels",
    random_state=0,
):
    np.random.seed(random_state)
    if groupby:
        if groupby not in adata.obs.columns:
            raise KeyError(f"{groupby} is not a valid obs annotation.")
        groups = adata.obs[groupby].unique()
        label_df = adata.obs[[groupby]].astype(str).rename(columns={groupby: key_added})
        for grp in groups:
            k = adata.obs[groupby] == grp
            grp_size = sum(k)
            n_partition = max(np.round(grp_size / partition_size).astype(int), 1)
            if method == "random":
                part_idx = np.random.randint(low=0, high=n_partition, size=grp_size)
            elif method == "random_even":
                part_sizes = list(map(len, np.array_split(np.arange(grp_size), n_partition)))
                part_idx = np.repeat(np.arange(n_partition), part_sizes)
                np.random.shuffle(part_idx)
            else:
                raise NotImplementedError(method)
            label_df.loc[k, key_added] = [f"{grp},{i}" for i in part_idx]
        adata.obs[key_added] = label_df[key_added]
    else:
        n_partition = max(np.round(adata.n_obs / partition_size).astype(int), 1)
        if method == "random":
            part_idx = np.random.randint(low=0, high=n_partition, size=adata.n_obs)
        elif method == "random_even":
            part_sizes = list(map(len, np.array_split(np.arange(adata.n_obs), n_partition)))
            part_idx = np.repeat(np.arange(n_partition), part_sizes)
            np.random.shuffle(part_idx)
        else:
            raise NotImplementedError(method)
        adata.obs[key_added] = part_idx.astype(str)


def dummy_to_categorical(mat, random_state=0):
    """Convert a sparse dummy matrix into a list of category indices, when a row
    has multiple entries, randomly assign to one of them.

    *Parameters
        + mat: csr_matrix
        A csr_matrix of ones, with cells on the rows and nhoods on the columns
        + random_state: int
        Seed for numpy RNG
    """
    np.random.seed(random_state)
    nrow = mat.shape[0]
    nhoods = []
    for i in range(nrow):
        k = (mat[i, ] == 1).indices
        if k.size == 1:
            idx = k[0]
        elif k.size > 1:
            idx = np.random.choice(k, 1)[0]
        else:
            idx = -1
        nhoods.append(idx)
    return nhoods


def pseudo_bulk(adata, groupby, use_rep="X", FUN=np.mean):
    """Make pseudo bulk data from grouped sc data"""
    grouping = adata.obs[groupby]
    if grouping.dtype.name != "category":
        grouping = grouping.astype("category")
    group_values = grouping.values
    groups = grouping.cat.categories.values
    n_group = len(groups)

    x, features = _find_rep(adata, use_rep)

    if sp.issparse(x):
        summarised = np.zeros((n_group, x.shape[1]))
        for i, grp in enumerate(groups):
            k_grp = group_values == grp
            summarised[i] = FUN(x[k_grp, :], axis=0)
    else:
        summarised = npg.aggregate_np(grouping.cat.codes, x, FUN, axis=0)
    return pd.DataFrame(summarised.T, columns=groups, index=features)


def summarise_expression_by_group(adata, groupby, genes=None, use_rep="X"):
    grouping = adata.obs[groupby]
    if grouping.dtype.name != "category":
        grouping = grouping.astype("category")
    group_values = grouping.values
    groups = grouping.cat.categories.values
    n_group = len(groups)

    x, features = _find_rep(adata, use_rep)
    if genes is not None:
        k_gene = np.in1d(features, genes).nonzero()[0]
        x = x[:, k_gene]
        features = features[k_gene]
    n_gene = features.size

    if sp.issparse(x):
        summarised_mean = np.zeros((n_group, n_gene))
        summarised_frac = np.zeros((n_group, n_gene))
        for i, grp in enumerate(groups):
            k_grp = group_values == grp
            summarised_mean[i] = np.mean(x[k_grp, :], axis=0)
            summarised_frac[i] = np.mean(x[k_grp, :] > 0, axis=0)
    else:
        summarised_mean = npg.aggregate_np(grouping.cat.codes, x, np.mean, axis=0)
        summarised_frac = npg.aggregate_np(grouping.cat.codes, x > 0, np.mean, axis=0)
    return pd.DataFrame(
        {
            "gene": np.repeat(features, n_group),
            "group": np.tile(groups, n_gene),
            "frac": summarised_frac.T.flatten(),
            "avg": summarised_mean.T.flatten(),
        }
    )


def score_msigdb_genesets(
    adata, groupby, msigdb, use_rep="raw", pattern=r"^HALLMARK_", use_pseudobulk=False, **pb_kw
):
    def read_msigdb(msigdb, pattern):
        genesets = {}
        if isinstance(pattern, str):
            pattern = re.compile(pattern)
        with open(msigdb, "r", encoding="utf8") as fh:
            for line in fh:
                fields = line.rstrip().split("\t")
                geneset_name = fields[0]
                geneset = fields[2:]
                if re.search(pattern, geneset_name):
                    genesets[geneset_name] = geneset
        print(f"read {len(genesets)} genesets")
        return genesets

    genesets = read_msigdb(msigdb, pattern)
    if use_pseudobulk:
        X = pseudo_bulk(adata, groupby, use_rep=use_rep, **pb_kw).T
        ad = anndata.AnnData(X=X)
    else:
        X = _find_rep(adata, use_rep, return_var_names=False)
        ad = anndata.AnnData(X, obs=adata.obs[[groupby]])

    for geneset_name, geneset in genesets.items():
        sc.tl.score_genes(
            ad,
            geneset,
            ctrl_size=max(50, len(geneset)),
            use_raw=(use_rep == "raw"),
            score_name=geneset_name,
        )

    del genesets

    return ad


def write_cellxgene_object(adata, output, **kwargs):
    ad = restore_adata(adata, restore_type="norm", use_raw=adata.raw, **kwargs)
    ad.write(output, compression="lzf")
    return ad



"""
scanpy umap
"""

import scanpy as sc
from ._obj_utils import (
    _set_default_key,
    _restore_default_key,
    _backup_obsm_key,
    _rename_obsm_key,
    _delete_obsm_backup_key,
    write_embedding,
)


def umap(
    adata,
    method=None,
    use_graph="neighbors",
    key_added=None,
    random_state=0,
    export_embedding=None,
    **kwargs,
):
    """
    Wrapper function for sc.tl.umap, for supporting named slot of umap embeddings
    """
    if sc.__version__.startswith("1.4"):
        _set_default_key(adata.uns, "neighbors", use_graph)
    if method == "rapids":
        kwargs["method"] = "rapids"
    if not isinstance(random_state, (list, tuple)):
        _backup_obsm_key(adata, "X_umap")

        if sc.__version__.startswith("1.4"):
            sc.tl.umap(adata, random_state=random_state, **kwargs)
        else:
            sc.tl.umap(adata, neighbors_key=use_graph, random_state=random_state, **kwargs)

        umap_key = "X_umap"
        if key_added:
            umap_key = f"X_umap_{key_added}"
            _rename_obsm_key(adata, "X_umap", umap_key)
        else:
            _delete_obsm_backup_key(adata, "X_umap")

        if export_embedding is not None:
            write_embedding(adata, umap_key, export_embedding, key_added=key_added)
    else:
        for i, rseed in enumerate(random_state):
            if key_added is None:
                umap_key = f"r{rseed}"
            elif not isinstance(key_added, (list, tuple)):
                umap_key = f"{key_added}_r{rseed}"
            elif len(key_added) == len(random_state):
                umap_key = key_added[i]
            else:
                raise ValueError(
                    "`key_added` can only be None, a scalar, or an "
                    "iterable of the same length as `random_state`."
                )
            umap(
                adata,
                use_graph="neighbors",
                key_added=umap_key,
                random_state=rseed,
                **kwargs,
            )
    if sc.__version__.startswith("1.4"):
        _restore_default_key(adata.uns, "neighbors", use_graph)



"""
Provides exported functions
"""

from ._read import (
    read_10x,
    read_10x_atac,
    read_h5ad,
    read_cellbender,
    read_velocyto,
)
from ._filter import filter_anndata
from ._norm import normalize
from ._hvg import hvg
from ._highly_deviant_genes import highly_deviant_genes
from ._pca import pca
from ._neighbors import neighbors
from ._umap import umap
from ._fdg import fdg
from ._tsne import tsne
from ._louvain import louvain
from ._leiden import leiden, leiden_shredding, leiden_shredding2
from ._diffexp import diffexp, diffexp_paired, extract_de_table
from ._diffmap import diffmap
from ._dpt import dpt
from ._paga import paga, plot_paga
from ._doublets import run_scrublet
from ._markers import (
    volcano_plot,
    calc_marker_stats,
    filter_marker_stats,
    top_markers,
    test_markers,
)
from ._utils import (
    cross_table,
    dummy_to_categorical,
    expand_feature_space,
    find_top_expressed_genes,
    lognorm_to_counts,
    project_into_PC,
    pseudo_bulk,
    random_partition,
    read_list,
    regroup,
    remove_genes,
    restore_adata,
    run_bbknn,
    run_cNMF,
    run_cellphonedb,
    run_celltype_composition_analysis,
    run_celltypist,
    run_diffQuant,
    run_harmony,
    run_phate,
    sc_warn,
    score_msigdb_genesets,
    show_obs_categories,
    split_by_group,
    subsample,
    summarise_expression_by_group,
    write_10x_h5,
    write_cellxgene_object,
    write_mtx,
    write_table,
)
from ._plot import (
    abline,
    clear_colors,
    dotplot2,
    dotplot3,
    dotplot_combined_coexpression,
    expression_colormap,
    heatmap,
    highlight,
    plot_cellbender_qc,
    plot_composition,
    plot_df_heatmap,
    plot_diffexp,
    plot_embedding,
    plot_genes,
    plot_markers,
    plot_metric_by_rank,
    plot_qc_scatter,
    plot_qc_violin,
    plot_scatter,
    set_figsize,
)
from ._annot import (
    LR_predict,
    LR_train,
    annotate,
    get_top_LR_features,
)
from ._velocity import run_scvelo, plot_scvelo
from ._pipeline import (
    QcLowPassError,
    auto_filter_cells,
    auto_zoom_in,
    calculate_qc,
    cellwise_qc,
    clusterwise_qc,
    crossmap,
    custom_pipeline,
    generate_qc_clusters,
    get_good_sized_batch,
    integrate,
    recluster_subset,
    simple_default_pipeline,
)



"""
scanpy norm
"""

import scanpy as sc


def normalize(adata, save_raw="yes", log_transform=True, **kwargs):
    """
    Wrapper function for sc.pp.normalize_per_cell() and sc.pp.log1p(), mainly
    for supporting different ways of saving raw data.
    """
    if save_raw == "counts":
        adata.raw = adata
    sc.pp.normalize_total(adata, **kwargs)
    if log_transform:
        sc.pp.log1p(adata)
    if save_raw == "yes":
        adata.raw = adata



"""
Provides read_10x()
"""

import os
from tempfile import mkdtemp
import numpy as np
import pandas as pd
import scanpy as sc
import h5py
import anndata


def read_10x(
    input_10x_h5,
    input_10x_mtx,
    genome="hg19",
    var_names="gene_symbols",
    extra_obs=None,
    extra_var=None,
):
    """
    Wrapper function for sc.read_10x_h5() and sc.read_10x_mtx(), mainly to
    support adding extra metadata
    """
    if input_10x_h5 is not None:
        adata = sc.read_10x_h5(input_10x_h5, genome=genome)
    elif input_10x_mtx is not None:
        adata = sc.read_10x_mtx(input_10x_mtx, var_names=var_names)

    if extra_obs:
        obs_tbl = pd.read_csv(extra_obs, sep="\t", header=0, index_col=0)
        adata.obs = adata.obs.merge(
            obs_tbl,
            how="left",
            left_index=True,
            right_index=True,
            suffixes=(False, False),
        )

    if extra_var:
        var_tbl = pd.read_csv(extra_var, sep="\t", header=0, index_col=0)
        adata.var = adata.var.merge(
            var_tbl,
            how="left",
            left_index=True,
            right_index=True,
            suffixes=(False, False),
        )
    return adata


def read_10x_atac(input_10x_mtx, extra_obs=None, extra_var=None):
    """
    Wrapper function for sc.read_10x_mtx() to read 10x ATAC data
    """
    matrix_mtx = os.path.join(input_10x_mtx, "matrix.mtx")
    barcodes_tsv = os.path.join(input_10x_mtx, "barcodes.tsv")
    peaks_bed = os.path.join(input_10x_mtx, "peaks.bed")
    if not os.path.exists(peaks_bed):
        raise FileNotFoundError

    tmp_dir = mkdtemp()
    genes_tsv = os.path.join(tmp_dir, "genes.tsv")

    os.symlink(os.path.abspath(matrix_mtx), os.path.join(tmp_dir, "matrix.mtx"))
    os.symlink(os.path.abspath(barcodes_tsv), os.path.join(tmp_dir, "barcodes.tsv"))
    with open(peaks_bed) as f, open(genes_tsv, "w") as fw:
        for line in f:
            fields = line.rstrip().split("\t")
            peak_id = "_".join(fields)
            print(f"{peak_id}\t{peak_id}", file=fw)

    adata = sc.read_10x_mtx(tmp_dir)
    adata.var.rename(columns={"gene_ids": "peak"}, inplace=True)

    os.remove(os.path.join(tmp_dir, "matrix.mtx"))
    os.remove(os.path.join(tmp_dir, "barcodes.tsv"))
    os.remove(genes_tsv)
    os.removedirs(tmp_dir)

    if extra_obs:
        obs_tbl = pd.read_csv(extra_obs, sep="\t", header=0, index_col=0)
        adata.obs = adata.obs.merge(
            obs_tbl,
            how="left",
            left_index=True,
            right_index=True,
            suffixes=(False, False),
        )

    if extra_var:
        var_tbl = pd.read_csv(extra_var, sep="\t", header=0, index_col=0)
        adata.var = adata.var.merge(
            var_tbl,
            how="left",
            left_index=True,
            right_index=True,
            suffixes=(False, False),
        )
    return adata


def read_cellbender(
    input_h5,
    remove_zero=True,
    remove_nan=True,
    train_history=False,
    latent_gene_encoding=False,
    add_suffix=None,
):
    """
    Read cellbender output h5 generated from mtx input
    """
    import scipy.sparse as sp

    f = h5py.File(input_h5, "r")
    if "matrix" in f:
        mat = f["matrix"]
        feat = mat["features"]
        feat_name = feat["name"][()]
        vardict = {
            "gene_ids": feat["id"][()].astype(str),
            "feature_type": feat["feature_type"][:].astype(str),
        }
    elif "background_removed" in f:
        mat = f["background_removed"]
        vardict = {
            "gene_ids": mat["genes"][()].astype(str),
        }
        feat_name = mat["gene_names"][()]
    else:
        raise ValueError("The data doesn't look like cellbender output")
    n_var, n_obs = tuple(mat["shape"][()])
    cols = ["latent_cell_probability", "latent_RT_efficiency"]
    if "barcode_indices_for_latents" in mat:
        bidx = mat["barcode_indices_for_latents"][()]
        obsdict = {}
        for x in cols:
            val = np.empty(n_obs)
            val.fill(np.nan)
            val[bidx] = mat[x][()]
            obsdict[x] = val
        if latent_gene_encoding:
            lge = mat["latent_gene_encoding"][()]
            obsm = np.empty((n_obs, lge.shape[1]))
            obsm.fill(np.nan)
            obsm[bidx, :] = lge
    else:
        obsdict = {x: mat[x] for x in cols}
        if latent_gene_encoding:
            obsm = mat["latent_gene_encoding"][()]
    barcodes = np.array(
        [b[:-2] if b.endswith("-1") else b for b in mat["barcodes"][()].astype(str)]
    )
    ad = anndata.AnnData(
        X=sp.csr_matrix(
            (mat["data"][()], mat["indices"][()], mat["indptr"][()]),
            shape=(n_obs, n_var),
        ),
        var=pd.DataFrame(vardict, index=feat_name.astype(str)),
        obs=pd.DataFrame(obsdict, index=barcodes),
        uns={
            "target_false_positive_rate": mat["target_false_positive_rate"][()],
            "test_elbo": list(mat["test_elbo"]),
            "test_epoch": list(mat["test_epoch"]),
            "training_elbo_per_epoch": list(mat["training_elbo_per_epoch"]),
        }
        if train_history
        else {},
    )
    ad.var_names_make_unique()
    if latent_gene_encoding:
        ad.obsm["X_latent_gene_encoding"] = obsm

    mask_nan = np.isnan(ad.obs.latent_cell_probability)
    mask_0 = ad.X.sum(axis=1).A1 <= 0

    mask_remove = np.zeros(n_obs).astype(bool)
    if remove_nan:
        mask_remove = mask_remove | mask_nan
    if remove_zero:
        mask_remove = mask_remove | mask_0
    idx_remove = np.where(mask_remove)[0]

    idx_sort = pd.Series(np.argsort(ad.obs_names))
    idx_sort = idx_sort[~idx_sort.isin(idx_remove)]

    ad1 = ad[idx_sort.values].copy()
    del ad

    if add_suffix:
        ad1.obs_names = ad1.obs_names.astype(str) + add_suffix

    return ad1


def read_h5ad(input_h5ad, component="all", **kwargs):
    if component == "all":
        return sc.read(input_h5ad, **kwargs)
    elif component == "raw":
        with h5py.File(input_h5ad, "r") as f:
            raw_dict = anndata._io.utils.read_attribute(f["/raw"])
            obs = anndata._io.h5ad.read_dataframe(f["/obs"])
            return anndata.AnnData(obs=obs, **raw_dict)
    elif isinstance(component, str):
        component = "/" + component if not component.startswith("/") else component
        if component in {"/obs", "/var"}:
            with h5py.File(input_h5ad, "r") as f:
                df = anndata._io.h5ad.read_dataframe(f[component])
                if not isinstance(f[component], h5py.Group):
                    for cats_name in f["/uns"].keys():
                        if not cats_name.endswith("_categories"):
                            continue
                        name = cats_name.replace("_categories", "")
                        cats = f[f"/uns/{cats_name}"][:]
                        if isinstance(cats, (str, int)):
                            cats = [cats]
                        if name in df.columns:
                            df[name] = pd.Categorical.from_codes(df[name].values, cats)
                return df
        elif component.startswith("/uns/neighbors") and len(component.split("/")) == 3:
            with h5py.File(input_h5ad, "r") as f:
                neighbor = anndata._io.h5ad.read_attribute(f[component])
                if "connectivities_key" in neighbor.keys():
                    c_key = neighbor["connectivities_key"]
                    neighbor["connectivities"] = anndata._io.h5ad.read_attribute(
                        f[f"/obsp/{c_key}"]
                    )
                if "distances_key" in neighbor.keys():
                    d_key = neighbor["distances_key"]
                    neighbor["distances"] = anndata._io.h5ad.read_attribute(f[f"/obsp/{d_key}"])
                return neighbor
        else:
            with h5py.File(input_h5ad, "r") as f:
                return anndata._io.utils.read_attribute(f[component])
    elif isinstance(component, (list, tuple)):
        d = dict()
        for comp in component:
            d[comp] = read_h5ad(input_h5ad, component=comp, **kwargs)
        return d
    else:
        raise ValueError(
            "`component` must be a str ('all', 'raw' or absolute HDF5 path), tuple or list of str"
        )


def read_velocyto(
    mtx_path,
    spliced_mtx="spliced.mtx.gz",
    unspliced_mtx="unspliced.mtx.gz",
    ambiguous_mtx="ambiguous.mtx.gz",
):

    barcodes_tsv = os.path.join(mtx_path, "barcodes.tsv.gz")
    features_tsv = os.path.join(mtx_path, "features.tsv.gz")
    mtx_dict = {"spliced": spliced_mtx, "unspliced": unspliced_mtx, "ambiguous": ambiguous_mtx}
    ad_dict = {}
    for name, mtx in mtx_dict.items():
        matrix_mtx = os.path.join(mtx_path, mtx)
        tmp_dir = mkdtemp()
        tmp_barcodes_tsv = os.path.join(tmp_dir, "barcodes.tsv.gz")
        tmp_features_tsv = os.path.join(tmp_dir, "features.tsv.gz")
        tmp_matrix_mtx = os.path.join(tmp_dir, "matrix.mtx.gz")
        os.symlink(barcodes_tsv, tmp_barcodes_tsv)
        os.symlink(features_tsv, tmp_features_tsv)
        os.symlink(matrix_mtx, tmp_matrix_mtx)
        try:
            ad_dict[name] = sc.read_10x_mtx(tmp_dir)
        finally:
            os.remove(tmp_barcodes_tsv)
            os.remove(tmp_features_tsv)
            os.remove(tmp_matrix_mtx)
            os.rmdir(tmp_dir)
    spliced_ad = ad_dict["spliced"]
    spliced_ad.layers["unspliced"] = ad_dict["unspliced"].X.copy()
    spliced_ad.layers["ambiguous"] = ad_dict["ambiguous"].X.copy()
    del ad_dict["unspliced"], ad_dict["ambiguous"]
    return spliced_ad



import scanpy as sc
from ._pca import pca
from ._plot import clear_colors, plot_embedding
from ._utils import remove_genes, run_harmony

try:
    import scvelo as scv

    def run_scvelo(
        adata, groupby, groups=None, batch=None, mode="dynamical", use_rep="X_pca", pca_kw={}
    ):
        ad = remove_genes(adata, ["mito", "ribo", "hb"])
        scv.pp.filter_and_normalize(ad, min_shared_counts=20, n_top_genes=2000)
        pca(ad, **pca_kw)
        # sc.pp.pca(ad)
        if batch:
            run_harmony(ad, batch=batch, key_added="hm")
        scv.pp.moments(ad, use_rep=use_rep, n_neighbors=30, n_pcs=30)
        if mode == "dynamical":
            scv.tl.recover_dynamics(ad)
        scv.tl.velocity(ad, mode=mode, diff_kinetics=(mode == "dynamical"))
        scv.tl.velocity_graph(ad)
        scv.tl.latent_time(ad)
        scv.tl.velocity_pseudotime(ad)
        # scv.tl.velocity(ad, mode='deterministic', vkey='velocity_det', groupby=groupby, groups=groups, filter_genes=False)
        # scv.tl.velocity_graph(ad, vkey='velocity_det')
        # scv.tl.velocity(ad, mode='stochastic', vkey='velocity_sto', groupby=groupby, groups=groups, filter_genes=False)
        # scv.tl.velocity_graph(ad, vkey='velocity_sto')
        # scv.tl.recover_dynamics(ad)
        # scv.tl.velocity(ad, mode='dynamical', vkey='velocity_dyn', groupby=groupby, groups=groups, filter_genes=False)
        # scv.tl.velocity_graph(ad, vkey='velocity_dyn')
        # scv.tl.terminal_states(ad, vkey='velocity_dyn', groupby=groupby, groups=groups)
        # scv.tl.latent_time(ad, vkey='velocity_dyn')
        # scv.tl.velocity_pseudotime(ad, vkey='velocity_dyn')
        return ad

    def plot_scvelo(
        adata, basis, groupby=None, groups=None, smooth=1.5, figsize=(4, 4), alpha=0.3, palette=None
    ):
        if groups:
            ad = adata[adata.obs[groupby].isin(groups)].copy()
        else:
            ad = adata
        if palette is None:
            clear_colors(ad)
        scv_plot_settings = {
            "figsize": figsize,
            "size": 100,
            "palette": palette,
            "alpha": alpha,
            "legend_loc": "right margin",
            "legend_fontsize": 15,
        }
        scv.pl.proportions(ad, groupby=groupby)
        plot_embedding(
            ad, basis=basis, groupby=groupby, annot="full", figsize=figsize, palette=palette
        )
        scv.pl.velocity_embedding_stream(ad, basis=basis, color=groupby)
        # scv.pl.velocity_embedding_stream(
        #    ad, vkey='velocity_det', basis=basis, color=groupby, smooth=smooth, title='deterministic velocity', **scv_plot_settings)
        # scv.pl.velocity_embedding_stream(
        #    ad, vkey='velocity_sto', basis=basis, color=groupby, smooth=smooth, title='stochastic velocity', **scv_plot_settings)
        # scv.pl.velocity_embedding_stream(
        #    ad, vkey='velocity_dyn', basis=basis, color=groupby, smooth=smooth, title='dynamic velocity', **scv_plot_settings)
        scv.pl.scatter(ad, basis=basis, c=("root_cells", "end_points", "latent_time"), size=30)

except ImportError:

    def run_scvelo(*args, **kwargs):
        raise NotImplementedError

    def plot_scvelo(*args, **kwargs):
        raise NotImplementedError
